# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Fetches the Grid Frequency."""

from __future__ import annotations

import asyncio
import logging

from frequenz.channels import Receiver, Sender
from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Frequency, Quantity

from .._internal._channels import ChannelRegistry
from ..microgrid import connection_manager
from ..microgrid._data_sourcing import ComponentMetricRequest
from ..timeseries._base_types import Sample

_logger = logging.getLogger(__name__)


def create_request(component_id: int) -> ComponentMetricRequest:
    """Create a request for grid frequency.

    Args:
        component_id: The component id to use for the request.

    Returns:
        A component metric request for grid frequency.
    """
    return ComponentMetricRequest(
        "grid-frequency", component_id, ComponentMetricId.FREQUENCY, None
    )


class GridFrequency:
    """Grid Frequency."""

    def __init__(
        self,
        data_sourcing_request_sender: Sender[ComponentMetricRequest],
        channel_registry: ChannelRegistry,
        source: Component | None = None,
    ):
        """Initialize the grid frequency formula generator.

        Args:
            data_sourcing_request_sender: The sender to use for requests.
            channel_registry: The channel registry to use for the grid frequency.
            source: The source component to use to receive the grid frequency.
        """
        if not source:
            component_graph = connection_manager.get().component_graph
            source = component_graph.find_first_descendant_component(
                root_category=ComponentCategory.GRID,
                descendant_categories=(
                    ComponentCategory.METER,
                    ComponentCategory.INVERTER,
                    ComponentCategory.EV_CHARGER,
                ),
            )

        self._request_sender: Sender[ComponentMetricRequest] = (
            data_sourcing_request_sender
        )
        self._channel_registry: ChannelRegistry = channel_registry
        self._source_component: Component = source
        self._component_metric_request: ComponentMetricRequest = create_request(
            self._source_component.component_id
        )

        self._task: None | asyncio.Task[None] = None

    @property
    def source(self) -> Component:
        """The component that is used to fetch the grid frequency.

        Returns:
            The component that is used for grid frequency.
        """
        return self._source_component

    def new_receiver(self) -> Receiver[Sample[Frequency]]:
        """Create a receiver for grid frequency.

        Returns:
            A receiver that will receive grid frequency samples.
        """
        receiver = self._channel_registry.get_or_create(
            Sample[Quantity], self._component_metric_request.get_channel_name()
        ).new_receiver()

        if not self._task:
            self._task = asyncio.create_task(self._send_request())
        else:
            _logger.info(
                "Grid frequency request already sent: %s", self._source_component
            )

        return receiver.map(
            lambda sample: (
                Sample[Frequency](sample.timestamp, None)
                if sample.value is None or sample.value.isnan()
                else Sample(
                    sample.timestamp, Frequency.from_hertz(sample.value.base_value)
                )
            )
        )

    async def _send_request(self) -> None:
        """Send the request for grid frequency."""
        await self._request_sender.send(self._component_metric_request)
        _logger.debug("Sent request for grid frequency: %s", self._source_component)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A logical meter for calculating high level metrics for a microgrid."""


import uuid

from frequenz.channels import Sender
from frequenz.quantities import Power, Quantity

from ..._internal._channels import ChannelRegistry
from ...microgrid._data_sourcing import ComponentMetricId, ComponentMetricRequest
from ..formula_engine import FormulaEngine
from ..formula_engine._formula_engine_pool import FormulaEnginePool
from ..formula_engine._formula_generators import CHPPowerFormula


class LogicalMeter:
    """A logical meter for calculating high level metrics in a microgrid.

    LogicalMeter provides methods for fetching power values from different points in the
    microgrid.  These methods return `FormulaReceiver` objects, which can be used like
    normal `Receiver`s, but can also be composed to form higher-order formula streams.

    !!! note
        `LogicalMeter` instances are not meant to be created directly by users.  Use the
        [`microgrid.logical_meter`][frequenz.sdk.microgrid.logical_meter] method for
        creating `LogicalMeter` instances.

    Example:
        ```python
        from datetime import timedelta

        from frequenz.sdk import microgrid
        from frequenz.sdk.actor import ResamplerConfig
        from frequenz.client.microgrid import ComponentMetricId


        await microgrid.initialize(
            "grpc://microgrid.sandbox.api.frequenz.io:62060",
            ResamplerConfig(resampling_period=timedelta(seconds=1)),
        )

        logical_meter = (
            microgrid.logical_meter()
            .start_formula("#1001 + #1002", ComponentMetricId.ACTIVE_POWER)
            .new_receiver()
        )

        async for power in logical_meter:
            print(power.value)
        ```
    """

    def __init__(
        self,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
    ) -> None:
        """Create a `LogicalMeter` instance.

        !!! note
            `LogicalMeter` instances are not meant to be created directly by users.  Use
            the [`microgrid.logical_meter`][frequenz.sdk.microgrid.logical_meter] method
            for creating `LogicalMeter` instances.

        Args:
            channel_registry: A channel registry instance shared with the resampling
                actor.
            resampler_subscription_sender: A sender for sending metric requests to the
                resampling actor.
        """
        self._channel_registry: ChannelRegistry = channel_registry
        self._resampler_subscription_sender: Sender[ComponentMetricRequest] = (
            resampler_subscription_sender
        )

        # Use a randomly generated uuid to create a unique namespace name for the local
        # meter to use when communicating with the resampling actor.
        self._namespace = f"logical-meter-{uuid.uuid4()}"
        self._formula_pool = FormulaEnginePool(
            self._namespace,
            self._channel_registry,
            self._resampler_subscription_sender,
        )

    def start_formula(
        self,
        formula: str,
        component_metric_id: ComponentMetricId,
        *,
        nones_are_zeros: bool = False,
    ) -> FormulaEngine[Quantity]:
        """Start execution of the given formula.

        Formulas can have Component IDs that are preceeded by a pound symbol("#"), and
        these operators: +, -, *, /, (, ).

        For example, the input string: "#20 + #5" is a formula for adding metrics from
        two components with ids 20 and 5.

        Args:
            formula: formula to execute.
            component_metric_id: The metric ID to use when fetching receivers from the
                resampling actor.
            nones_are_zeros: Whether to treat None values from the stream as 0s.  If
                False, the returned value will be a None.

        Returns:
            A FormulaEngine that applies the formula and streams values.
        """
        return self._formula_pool.from_string(
            formula, component_metric_id, nones_are_zeros=nones_are_zeros
        )

    @property
    def chp_power(self) -> FormulaEngine[Power]:
        """Fetch the CHP power production in the microgrid.

        This formula produces values that are in the Passive Sign Convention (PSC).

        If a formula engine to calculate CHP power production is not already running, it
        will be started.

        A receiver from the formula engine can be created using the `new_receiver`
        method.

        Returns:
            A FormulaEngine that will calculate and stream CHP power production.
        """
        engine = self._formula_pool.from_power_formula_generator(
            "chp_power",
            CHPPowerFormula,
        )
        assert isinstance(engine, FormulaEngine)
        return engine

    async def stop(self) -> None:
        """Stop all formula engines."""
        await self._formula_pool.stop()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Actor model implementation."""

import abc
import asyncio
import logging
from datetime import timedelta

from ._background_service import BackgroundService

_logger = logging.getLogger(__name__)


class Actor(BackgroundService, abc.ABC):
    """A primitive unit of computation that runs autonomously.

    To implement an actor, subclasses must implement the
    [`_run()`][frequenz.sdk.actor--the-_run-method] method, which should run the actor's
    logic. The [`_run()`][frequenz.sdk.actor--the-_run-method] method is called by the
    base class when the actor is started, and is expected to run until the actor is
    stopped.

    !!! info

        Please read the [`actor` module documentation][frequenz.sdk.actor] for more
        comprehensive guide on how to use and implement actors properly.
    """

    RESTART_DELAY: timedelta = timedelta(seconds=2)
    """The delay to wait between restarts of this actor."""

    _restart_limit: int | None = None
    """The number of times actors can be restarted when they are stopped by unhandled exceptions.

    If this is bigger than 0 or `None`, the actor will be restarted when there is an
    unhanded exception in the `_run()` method.

    If `None`, the actor will be restarted an unlimited number of times.

    !!! note

        This is mostly used for testing purposes and shouldn't be set in production.
    """

    def start(self) -> None:
        """Start this actor.

        If this actor is already running, this method does nothing.
        """
        if self.is_running:
            return
        self._tasks.clear()
        self._tasks.add(asyncio.create_task(self._run_loop()))

    @abc.abstractmethod
    async def _run(self) -> None:
        """Run this actor's logic."""

    async def _delay_if_restart(self, iteration: int) -> None:
        """Delay the restart of this actor's n'th iteration.

        Args:
            iteration: The current iteration of the restart.
        """
        # NB: I think it makes sense (in the future) to think about deminishing returns
        # the longer the actor has been running.
        # Not just for the restart-delay but actually for the n_restarts counter as well.
        if iteration > 0:
            delay = self.RESTART_DELAY.total_seconds()
            _logger.info("Actor %s: Waiting %s seconds...", self, delay)
            await asyncio.sleep(delay)

    async def _run_loop(self) -> None:
        """Run the actor's task continuously, managing restarts, cancellation, and termination.

        This method handles the execution of the actor's task, including
        restarts for unhandled exceptions, cancellation, or normal termination.

        Raises:
            asyncio.CancelledError: If the actor's `_run()` method is cancelled.
            Exception: If the actor's `_run()` method raises any other exception.
            BaseException: If the actor's `_run()` method raises any base exception.
        """
        _logger.info("Actor %s: Started.", self)
        n_restarts = 0
        while True:
            try:
                await self._delay_if_restart(n_restarts)
                await self._run()
                _logger.info("Actor %s: _run() returned without error.", self)
            except asyncio.CancelledError:
                _logger.info("Actor %s: Cancelled.", self)
                raise
            except Exception:  # pylint: disable=broad-except
                _logger.exception("Actor %s: Raised an unhandled exception.", self)
                limit_str = "∞" if self._restart_limit is None else self._restart_limit
                limit_str = f"({n_restarts}/{limit_str})"
                if self._restart_limit is None or n_restarts < self._restart_limit:
                    n_restarts += 1
                    _logger.info("Actor %s: Restarting %s...", self._name, limit_str)
                    continue
                _logger.info(
                    "Actor %s: Maximum restarts attempted %s, bailing out...",
                    self,
                    limit_str,
                )
                raise
            except BaseException:  # pylint: disable=broad-except
                _logger.exception("Actor %s: Raised a BaseException.", self)
                raise
            break

        _logger.info("Actor %s: Stopped.", self)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Manage a pool of batteries."""

from ._battery_pool import BatteryPool
from .messages import BatteryPoolReport

__all__ = [
    "BatteryPool",
    "BatteryPoolReport",
]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""A formula pool for helping with tracking running formula engines."""

from __future__ import annotations

from typing import TYPE_CHECKING

from frequenz.channels import Sender
from frequenz.client.microgrid import ComponentMetricId
from frequenz.quantities import Current, Power, Quantity, ReactivePower

from ..._internal._channels import ChannelRegistry
from ...microgrid._data_sourcing import ComponentMetricRequest
from ._formula_generators._formula_generator import (
    FormulaGenerator,
    FormulaGeneratorConfig,
)
from ._resampled_formula_builder import ResampledFormulaBuilder

if TYPE_CHECKING:
    # Break circular import
    from ..formula_engine import FormulaEngine, FormulaEngine3Phase


class FormulaEnginePool:
    """Creates and owns formula engines from string formulas, or formula generators.

    If an engine already exists with a given name, it is reused instead.
    """

    def __init__(
        self,
        namespace: str,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
    ) -> None:
        """Create a new instance.

        Args:
            namespace: namespace to use with the data pipeline.
            channel_registry: A channel registry instance shared with the resampling
                actor.
            resampler_subscription_sender: A sender for sending metric requests to the
                resampling actor.
        """
        self._namespace: str = namespace
        self._channel_registry: ChannelRegistry = channel_registry
        self._resampler_subscription_sender: Sender[ComponentMetricRequest] = (
            resampler_subscription_sender
        )
        self._string_engines: dict[str, FormulaEngine[Quantity]] = {}
        self._power_engines: dict[str, FormulaEngine[Power]] = {}
        self._power_3_phase_engines: dict[str, FormulaEngine3Phase[Power]] = {}
        self._current_engines: dict[str, FormulaEngine3Phase[Current]] = {}
        self._reactive_power_engines: dict[str, FormulaEngine[ReactivePower]] = {}

    def from_string(
        self,
        formula: str,
        component_metric_id: ComponentMetricId,
        *,
        nones_are_zeros: bool = False,
    ) -> FormulaEngine[Quantity]:
        """Get a receiver for a manual formula.

        Args:
            formula: formula to execute.
            component_metric_id: The metric ID to use when fetching receivers from the
                resampling actor.
            nones_are_zeros: Whether to treat None values from the stream as 0s.  If
                False, the returned value will be a None.

        Returns:
            A FormulaReceiver that streams values with the formulas applied.
        """
        channel_key = formula + component_metric_id.value
        if channel_key in self._string_engines:
            return self._string_engines[channel_key]

        builder = ResampledFormulaBuilder(
            namespace=self._namespace,
            formula_name=formula,
            channel_registry=self._channel_registry,
            resampler_subscription_sender=self._resampler_subscription_sender,
            metric_id=component_metric_id,
            create_method=Quantity,
        )
        formula_engine = builder.from_string(formula, nones_are_zeros=nones_are_zeros)
        self._string_engines[channel_key] = formula_engine

        return formula_engine

    def from_reactive_power_formula_generator(
        self,
        channel_key: str,
        generator: type[FormulaGenerator[ReactivePower]],
        config: FormulaGeneratorConfig = FormulaGeneratorConfig(),
    ) -> FormulaEngine[ReactivePower]:
        """Get a receiver for a formula from a generator.

        Args:
            channel_key: A string to uniquely identify the formula.
            generator: A formula generator.
            config: config to initialize the formula generator with.

        Returns:
            A FormulaReceiver or a FormulaReceiver3Phase instance based on what the
                FormulaGenerator returns.
        """
        from ._formula_engine import (  # pylint: disable=import-outside-toplevel
            FormulaEngine,
        )

        if channel_key in self._reactive_power_engines:
            return self._reactive_power_engines[channel_key]

        engine = generator(
            self._namespace,
            self._channel_registry,
            self._resampler_subscription_sender,
            config,
        ).generate()
        assert isinstance(engine, FormulaEngine)
        self._reactive_power_engines[channel_key] = engine
        return engine

    def from_power_formula_generator(
        self,
        channel_key: str,
        generator: type[FormulaGenerator[Power]],
        config: FormulaGeneratorConfig = FormulaGeneratorConfig(),
    ) -> FormulaEngine[Power]:
        """Get a receiver for a formula from a generator.

        Args:
            channel_key: A string to uniquely identify the formula.
            generator: A formula generator.
            config: config to initialize the formula generator with.

        Returns:
            A FormulaReceiver or a FormulaReceiver3Phase instance based on what the
                FormulaGenerator returns.
        """
        from ._formula_engine import (  # pylint: disable=import-outside-toplevel
            FormulaEngine,
        )

        if channel_key in self._power_engines:
            return self._power_engines[channel_key]

        engine = generator(
            self._namespace,
            self._channel_registry,
            self._resampler_subscription_sender,
            config,
        ).generate()
        assert isinstance(engine, FormulaEngine)
        self._power_engines[channel_key] = engine
        return engine

    def from_power_3_phase_formula_generator(
        self,
        channel_key: str,
        generator: type[FormulaGenerator[Power]],
        config: FormulaGeneratorConfig = FormulaGeneratorConfig(),
    ) -> FormulaEngine3Phase[Power]:
        """Get a formula engine that streams 3-phase power values.

        Args:
            channel_key: The string to uniquely identify the formula.
            generator: The formula generator.
            config: The config to initialize the formula generator with.

        Returns:
            A formula engine that streams [3-phase][frequenz.sdk.timeseries.Sample3Phase]
            power values.
        """
        from ._formula_engine import (  # pylint: disable=import-outside-toplevel
            FormulaEngine3Phase,
        )

        if channel_key in self._power_3_phase_engines:
            return self._power_3_phase_engines[channel_key]

        engine = generator(
            self._namespace,
            self._channel_registry,
            self._resampler_subscription_sender,
            config,
        ).generate()
        assert isinstance(engine, FormulaEngine3Phase)
        self._power_3_phase_engines[channel_key] = engine
        return engine

    def from_3_phase_current_formula_generator(
        self,
        channel_key: str,
        generator: type[FormulaGenerator[Current]],
        config: FormulaGeneratorConfig = FormulaGeneratorConfig(),
    ) -> FormulaEngine3Phase[Current]:
        """Get a receiver for a formula from a generator.

        Args:
            channel_key: A string to uniquely identify the formula.
            generator: A formula generator.
            config: config to initialize the formula generator with.

        Returns:
            A FormulaReceiver or a FormulaReceiver3Phase instance based on what the
                FormulaGenerator returns.
        """
        from ._formula_engine import (  # pylint: disable=import-outside-toplevel
            FormulaEngine3Phase,
        )

        if channel_key in self._current_engines:
            return self._current_engines[channel_key]

        engine = generator(
            self._namespace,
            self._channel_registry,
            self._resampler_subscription_sender,
            config,
        ).generate()
        assert isinstance(engine, FormulaEngine3Phase)
        self._current_engines[channel_key] = engine
        return engine

    async def stop(self) -> None:
        """Stop all formula engines in the pool."""
        for string_engine in self._string_engines.values():
            await string_engine._stop()  # pylint: disable=protected-access
        for power_engine in self._power_engines.values():
            await power_engine._stop()  # pylint: disable=protected-access
        for power_3_phase_engine in self._power_3_phase_engines.values():
            await power_3_phase_engine._stop()  # pylint: disable=protected-access
        for current_engine in self._current_engines.values():
            await current_engine._stop()  # pylint: disable=protected-access
        for reactive_power_engine in self._reactive_power_engines.values():
            await reactive_power_engine._stop()  # pylint: disable=protected-access



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Methods for processing battery-inverter data."""


import asyncio
import logging
from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone
from typing import Generic

from frequenz.channels import Broadcast, Receiver

from ..._internal._asyncio import cancel_and_await, run_forever
from ..._internal._constants import RECEIVER_MAX_SIZE, WAIT_FOR_COMPONENT_DATA_SEC
from ...microgrid._power_distributing._component_managers._battery_manager import (
    _get_battery_inverter_mappings,
)
from ._component_metric_fetcher import (
    ComponentMetricFetcher,
    LatestBatteryMetricsFetcher,
    LatestInverterMetricsFetcher,
)
from ._component_metrics import ComponentMetricsData
from ._metric_calculator import MetricCalculator, T

_logger = logging.getLogger(__name__)


class MetricAggregator(Generic[T], ABC):
    """Interface to control how the component data should be aggregated and send."""

    @abstractmethod
    def update_working_batteries(self, new_working_batteries: set[int]) -> None:
        """Update set of the working batteries.

        Args:
            new_working_batteries: Set of the working batteries.
        """

    @abstractmethod
    def new_receiver(self, limit: int | None = RECEIVER_MAX_SIZE) -> Receiver[T]:
        """Return new receiver for the aggregated metric results.

        Args:
            limit: Buffer size of the receiver

        Returns:
            Receiver for the metric results.
        """

    @abstractmethod
    async def stop(self) -> None:
        """Stop the all pending async task."""

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        """Return name of this method.

        Returns:
            Name of this method.
        """


class SendOnUpdate(MetricAggregator[T]):
    """Wait for the change of the components metrics and send updated result.

    This method will cache the component metrics. When any metric change it will
    recalculate high level metric. If the calculation result change, it will
    send the new value.
    """

    def __init__(
        self,
        working_batteries: set[int],
        metric_calculator: MetricCalculator[T],
        min_update_interval: timedelta,
    ) -> None:
        """Create class instance.

        Args:
            working_batteries: Set of the working batteries
            metric_calculator: Module that tells how to aggregate the component metrics.
            min_update_interval: Minimum frequency for sending update about the change.
        """
        self._metric_calculator: MetricCalculator[T] = metric_calculator
        self._bat_inv_map = _get_battery_inverter_mappings(
            self._metric_calculator.batteries,
            inv_bats=False,
            bat_bats=False,
            inv_invs=False,
        )["bat_invs"]

        self._working_batteries: set[int] = working_batteries.intersection(
            metric_calculator.batteries
        )
        self._result_channel: Broadcast[T] = Broadcast(
            name=SendOnUpdate.name() + "_" + metric_calculator.name(),
            resend_latest=True,
        )

        self._update_event = asyncio.Event()
        self._cached_metrics: dict[int, ComponentMetricsData] = {}

        self._update_task = asyncio.create_task(run_forever(self._update_and_notify))
        self._send_task = asyncio.create_task(
            run_forever(lambda: self._send_on_update(min_update_interval))
        )
        self._pending_data_fetchers: set[asyncio.Task[ComponentMetricsData | None]] = (
            set()
        )

    @classmethod
    def name(cls) -> str:
        """Get name of the method.

        Returns:
            Name of the method.
        """
        return "SendOnUpdate"

    def new_receiver(self, limit: int | None = RECEIVER_MAX_SIZE) -> Receiver[T]:
        """Return new receiver for the aggregated metric results.

        Args:
            limit: Buffer size of the receiver

        Returns:
            Receiver for the metric results.
        """
        if limit is None:
            return self._result_channel.new_receiver()
        return self._result_channel.new_receiver(limit=limit)

    def update_working_batteries(self, new_working_batteries: set[int]) -> None:
        """Update set of the working batteries.

        Recalculate metric if set changed.

        Args:
            new_working_batteries: Set of the working batteries.
        """
        # MetricCalculator can discard some batteries if they doesn't meet requirements.
        # For example batteries without adjacent inverter won't be included
        # int he PowerBounds metrics.
        new_set = new_working_batteries.intersection(self._metric_calculator.batteries)

        stopped_working = self._working_batteries - new_set
        for battery_id in stopped_working:
            # Removed cached metrics for components that stopped working.
            self._cached_metrics.pop(battery_id, None)
            for inv_id in self._bat_inv_map[battery_id]:
                self._cached_metrics.pop(inv_id, None)

        if new_set != self._working_batteries:
            self._working_batteries = new_set
            self._update_event.set()

    async def stop(self) -> None:
        """Stop the all pending async task."""
        await asyncio.gather(
            *[cancel_and_await(task) for task in self._pending_data_fetchers]
        )
        await asyncio.gather(
            *[cancel_and_await(self._send_task), cancel_and_await(self._update_task)]
        )

    async def _create_data_fetchers(self) -> dict[int, ComponentMetricFetcher]:
        fetchers: dict[int, ComponentMetricFetcher] = {
            cid: await LatestBatteryMetricsFetcher.async_new(cid, metrics)
            for cid, metrics in self._metric_calculator.battery_metrics.items()
        }
        inverter_fetchers = {
            cid: await LatestInverterMetricsFetcher.async_new(cid, metrics)
            for cid, metrics in self._metric_calculator.inverter_metrics.items()
        }
        fetchers.update(inverter_fetchers)
        return fetchers

    def _remove_metric_fetcher(
        self, fetchers: dict[int, ComponentMetricFetcher], component_id: int
    ) -> None:
        _logger.error(
            "Removing component %d from the %s formula.",
            component_id,
            self._result_channel._name,  # pylint: disable=protected-access
        )
        fetchers.pop(component_id)

    def _metric_updated(self, new_metrics: ComponentMetricsData) -> bool:
        cid = new_metrics.component_id
        return (
            cid not in self._cached_metrics or new_metrics != self._cached_metrics[cid]
        )

    async def _update_and_notify(self) -> None:
        """Receive component metrics and send notification when they change."""
        fetchers = await self._create_data_fetchers()

        self._pending_data_fetchers = {
            asyncio.create_task(fetcher.fetch_next(), name=str(cid))
            for cid, fetcher in fetchers.items()
        }
        while len(self._pending_data_fetchers) > 0:
            done, self._pending_data_fetchers = await asyncio.wait(
                self._pending_data_fetchers, return_when=asyncio.FIRST_COMPLETED
            )
            for item in done:
                metrics = item.result()
                if metrics is None:
                    self._remove_metric_fetcher(fetchers, int(item.get_name()))
                    continue
                if self._metric_updated(metrics):
                    self._update_event.set()

                cid = metrics.component_id
                # Save metric even if not changed to update its timestamp.
                self._cached_metrics[cid] = metrics
                # Add fetcher back to the processing list.
                self._pending_data_fetchers.add(
                    asyncio.create_task(fetchers[cid].fetch_next(), name=str(cid))
                )

    async def _send_on_update(self, min_update_interval: timedelta) -> None:
        """Wait for an update notification, recalculate metric and send.

        If recalculated metric changed, then send new value.

        Args:
            min_update_interval: Minimum frequency for sending update about the change.
        """
        sender = self._result_channel.new_sender()
        latest_calculation_result: T | None = None
        # Wait for first data to come.
        # In that way the first outputs are reliable.
        await asyncio.sleep(WAIT_FOR_COMPONENT_DATA_SEC)

        while True:
            await self._update_event.wait()
            self._update_event.clear()

            result: T = self._metric_calculator.calculate(
                self._cached_metrics, self._working_batteries
            )
            if result != latest_calculation_result:
                latest_calculation_result = result
                await sender.send(result)

                if result is None:
                    sleep_for = min_update_interval.total_seconds()
                else:
                    # Sleep for the rest of the time.
                    # Then we won't send update more frequently then min_update_interval
                    time_diff = datetime.now(tz=timezone.utc) - result.timestamp
                    sleep_for = (min_update_interval - time_diff).total_seconds()
                await asyncio.sleep(sleep_for)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Module with constants shared between instances of the sdk.

To be replaced by ConfigManager.
"""

RECEIVER_MAX_SIZE = 50
"""Default buffer size of the receiver."""

WAIT_FOR_COMPONENT_DATA_SEC: float = 2
"""Delay the start of the application to wait for the data."""

MAX_BATTERY_DATA_AGE_SEC: float = 2
"""Max time difference for the battery or inverter data to be considered as reliable.

If battery or inverter stopped sending data, then this is the maximum time when its
last message should be considered as valid. After that time, component data
should not be used.
"""



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for Consumer Power."""

import logging

from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Power

from ....microgrid import connection_manager
from .._formula_engine import FormulaEngine
from .._resampled_formula_builder import ResampledFormulaBuilder
from ._fallback_formula_metric_fetcher import FallbackFormulaMetricFetcher
from ._formula_generator import (
    NON_EXISTING_COMPONENT_ID,
    ComponentNotFound,
    FormulaGenerator,
    FormulaGeneratorConfig,
)
from ._simple_formula import SimplePowerFormula

_logger = logging.getLogger(__name__)


class ConsumerPowerFormula(FormulaGenerator[Power]):
    """Formula generator from component graph for calculating the Consumer Power.

    The consumer power is calculated by summing up the power of all components that
    are not part of a battery, CHP, PV or EV charger chain.
    """

    def _are_grid_meters(self, grid_successors: set[Component]) -> bool:
        """Check if the grid successors are grid meters.

        Args:
            grid_successors: The successors of the grid component.

        Returns:
            True if the provided components are grid meters, False otherwise.
        """
        component_graph = connection_manager.get().component_graph
        return all(
            successor.category == ComponentCategory.METER
            and not component_graph.is_battery_chain(successor)
            and not component_graph.is_chp_chain(successor)
            and not component_graph.is_pv_chain(successor)
            and not component_graph.is_ev_charger_chain(successor)
            for successor in grid_successors
        )

    # We need the noqa here because `RuntimeError` is raised indirectly
    def generate(self) -> FormulaEngine[Power]:  # noqa: DOC503
        """Generate formula for calculating consumer power from the component graph.

        Returns:
            A formula engine that will calculate the consumer power.

        Raises:
            ComponentNotFound: If the component graph does not contain a consumer power
                component.
            RuntimeError: If the grid component has a single successor that is not a
                meter.
        """
        builder = self._get_builder(
            "consumer-power", ComponentMetricId.ACTIVE_POWER, Power.from_watts
        )

        grid_successors = self._get_grid_component_successors()

        if not grid_successors:
            raise ComponentNotFound("No components found in the component graph.")

        if self._are_grid_meters(grid_successors):
            return self._gen_with_grid_meter(builder, grid_successors)

        return self._gen_without_grid_meter(builder, self._get_grid_component())

    def _gen_with_grid_meter(
        self,
        builder: ResampledFormulaBuilder[Power],
        grid_meters: set[Component],
    ) -> FormulaEngine[Power]:
        """Generate formula for calculating consumer power with grid meter.

        Args:
            builder: The formula engine builder.
            grid_meters: The grid meter component.

        Returns:
            A formula engine that will calculate the consumer power.
        """
        assert grid_meters
        component_graph = connection_manager.get().component_graph

        def non_consumer_component(component: Component) -> bool:
            """
            Check if a component is not a consumer component.

            Args:
                component: The component to check.

            Returns:
                True if the component is not a consumer component, False otherwise.
            """
            # If the component graph supports additional types of grid successors in the
            # future, additional checks need to be added here.
            return (
                component_graph.is_battery_chain(component)
                or component_graph.is_chp_chain(component)
                or component_graph.is_pv_chain(component)
                or component_graph.is_ev_charger_chain(component)
            )

        # join all non consumer components reachable from the different grid meters
        non_consumer_components: set[Component] = set()
        for grid_meter in grid_meters:
            non_consumer_components = non_consumer_components.union(
                component_graph.dfs(grid_meter, set(), non_consumer_component)
            )

        # push all grid meters
        for idx, grid_meter in enumerate(grid_meters):
            if idx > 0:
                builder.push_oper("+")
            builder.push_component_metric(
                grid_meter.component_id, nones_are_zeros=False
            )

        if self._config.allow_fallback:
            fallbacks = self._get_fallback_formulas(non_consumer_components)

            for idx, (primary_component, fallback_formula) in enumerate(
                fallbacks.items()
            ):
                builder.push_oper("-")

                # should only be the case if the component is not a meter
                builder.push_component_metric(
                    primary_component.component_id,
                    nones_are_zeros=(
                        primary_component.category != ComponentCategory.METER
                    ),
                    fallback=fallback_formula,
                )
        else:
            # push all non consumer components and subtract them from the grid meters
            for component in non_consumer_components:
                builder.push_oper("-")
                builder.push_component_metric(
                    component.component_id,
                    nones_are_zeros=component.category != ComponentCategory.METER,
                )

        return builder.build()

    def _gen_without_grid_meter(
        self,
        builder: ResampledFormulaBuilder[Power],
        grid: Component,
    ) -> FormulaEngine[Power]:
        """Generate formula for calculating consumer power without a grid meter.

        Args:
            builder: The formula engine builder.
            grid: The grid component.

        Returns:
            A formula engine that will calculate the consumer power.
        """

        def consumer_component(component: Component) -> bool:
            """
            Check if a component is a consumer component.

            Args:
                component: The component to check.

            Returns:
                True if the component is a consumer component, False otherwise.
            """
            # If the component graph supports additional types of grid successors in the
            # future, additional checks need to be added here.
            return (
                component.category
                in {ComponentCategory.METER, ComponentCategory.INVERTER}
                and not component_graph.is_battery_chain(component)
                and not component_graph.is_chp_chain(component)
                and not component_graph.is_pv_chain(component)
                and not component_graph.is_ev_charger_chain(component)
            )

        component_graph = connection_manager.get().component_graph
        consumer_components = component_graph.dfs(grid, set(), consumer_component)

        if not consumer_components:
            _logger.warning(
                "Unable to find any consumers in the component graph. "
                "Subscribing to the resampling actor with a non-existing "
                "component id, so that `0` values are sent from the formula."
            )
            # If there are no consumer components, we have to send 0 values at the same
            # frequency as the other streams.  So we subscribe with a non-existing
            # component id, just to get a `None` message at the resampling interval.
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID, nones_are_zeros=True
            )
            return builder.build()

        if self._config.allow_fallback:
            fallbacks = self._get_fallback_formulas(consumer_components)

            for idx, (primary_component, fallback_formula) in enumerate(
                fallbacks.items()
            ):
                if idx > 0:
                    builder.push_oper("+")

                # should only be the case if the component is not a meter
                builder.push_component_metric(
                    primary_component.component_id,
                    nones_are_zeros=(
                        primary_component.category != ComponentCategory.METER
                    ),
                    fallback=fallback_formula,
                )
        else:
            for idx, component in enumerate(consumer_components):
                if idx > 0:
                    builder.push_oper("+")

                builder.push_component_metric(
                    component.component_id,
                    nones_are_zeros=component.category != ComponentCategory.METER,
                )

        return builder.build()

    def _get_fallback_formulas(
        self, components: set[Component]
    ) -> dict[Component, FallbackFormulaMetricFetcher[Power] | None]:
        """Find primary and fallback components and create fallback formulas.

        The primary component is the one that will be used to calculate the consumer power.
        However, if it is not available, the fallback formula will be used instead.
        Fallback formulas calculate the consumer power using the fallback components.
        Fallback formulas are wrapped in `FallbackFormulaMetricFetcher` to allow
        for lazy initialization.

        Args:
            components: The producer components.

        Returns:
            A dictionary mapping primary components to their FallbackFormulaMetricFetcher.
        """
        fallbacks = self._get_metric_fallback_components(components)

        fallback_formulas: dict[
            Component, FallbackFormulaMetricFetcher[Power] | None
        ] = {}

        for primary_component, fallback_components in fallbacks.items():
            if len(fallback_components) == 0:
                fallback_formulas[primary_component] = None
                continue

            fallback_ids = [c.component_id for c in fallback_components]
            generator = SimplePowerFormula(
                f"{self._namespace}_fallback_{fallback_ids}",
                self._channel_registry,
                self._resampler_subscription_sender,
                FormulaGeneratorConfig(
                    component_ids=set(fallback_ids),
                    allow_fallback=False,
                ),
            )

            fallback_formulas[primary_component] = FallbackFormulaMetricFetcher(
                generator
            )

        return fallback_formulas



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""The ComponentMetricRequest class."""

from dataclasses import dataclass
from datetime import datetime

from frequenz.client.microgrid import ComponentMetricId

__all__ = ["ComponentMetricRequest", "ComponentMetricId"]


@dataclass
class ComponentMetricRequest:
    """A request to start streaming a component's metric.

    Requesters use this class to specify which component's metric they want to subscribe
    to, including the component ID, metric ID, and an optional start time. The
    `namespace` is defined by the requester and influences the construction of the
    channel name via the `get_channel_name()` method.

    The `namespace` allows differentiation of data streams for the same component and
    metric. For example, requesters can use different `namespace` values to subscribe to
    raw or resampled data streams separately. This ensures that each requester receives
    the appropriate type of data without interference. Requests with the same
    `namespace`, `component_id`, and `metric_id` will use the same channel, preventing
    unnecessary duplication of data streams.

    The requester and provider must use the same channel name so that they can
    independently retrieve the same channel from the `ChannelRegistry`.  This is
    achieved by using the `get_channel_name` method to generate the name on both sides
    based on parameters set by the requesters.
    """

    namespace: str
    """A client-defined identifier influencing the channel name."""

    component_id: int
    """The ID of the requested component."""

    metric_id: ComponentMetricId
    """The ID of the requested component's metric."""

    start_time: datetime | None
    """The start time from which data is required.

    If None, only live data is streamed.
    """

    def get_channel_name(self) -> str:
        """Construct the channel name based on the request parameters.

        Returns:
            A string representing the channel name.
        """
        start = f",start={self.start_time}" if self.start_time else ""
        return (
            "component_metric_request<"
            f"namespace={self.namespace},"
            f"component_id={self.component_id},"
            f"metric_id={self.metric_id.name}"
            f"{start}"
            ">"
        )



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Ringbuffer implementation with focus on time & memory efficiency."""


from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Generic, SupportsIndex, TypeVar, overload

import numpy as np
import numpy.typing as npt

from .._base_types import UNIX_EPOCH, QuantityT, Sample

FloatArray = TypeVar("FloatArray", list[float], npt.NDArray[np.float64])
"""Type variable of the buffer container."""


@dataclass
class Gap:
    """A gap defines the range for which we haven't received values yet."""

    start: datetime
    """Start timestamp of the range, inclusive."""
    end: datetime
    """End timestamp of the range, exclusive."""

    def contains(self, timestamp: datetime) -> bool:
        """Check if a given timestamp is inside this gap.

        Args:
            timestamp: Timestamp to check.

        Returns:
            True if the timestamp is in the gap.
        """
        if self.start <= timestamp < self.end:
            return True

        return False


class OrderedRingBuffer(Generic[FloatArray]):
    """Time aware ringbuffer that keeps its entries sorted by time."""

    _TIMESTAMP_MIN = datetime.min.replace(tzinfo=timezone.utc)
    """The minimum representable timestamp."""

    _TIMESTAMP_MAX = datetime.max.replace(tzinfo=timezone.utc)
    """The maximum representable timestamp."""

    def __init__(
        self,
        buffer: FloatArray,
        sampling_period: timedelta,
        align_to: datetime = UNIX_EPOCH,
    ) -> None:
        """Initialize the time aware ringbuffer.

        Args:
            buffer: Instance of a buffer container to use internally.
            sampling_period: Timedelta of the desired sampling period.
            align_to: Arbitrary point in time used to align
                timestamped data with the index position in the buffer.
                Used to make the data stored in the buffer align with the
                beginning and end of the buffer borders.

                For example, if the `align_to` is set to
                "0001-01-01 12:00:00", and the `sampling_period` is set to
                1 hour and the length of the buffer is 24, then the data
                stored in the buffer could correspond to the time range from
                "2022-01-01 12:00:00" to "2022-01-02 12:00:00" (date chosen
                arbitrarily here).
        """
        assert len(buffer) > 0, "The buffer capacity must be higher than zero"

        self._buffer: FloatArray = buffer
        self._sampling_period: timedelta = sampling_period
        self._time_index_alignment: datetime = align_to

        self._gaps: list[Gap] = []
        self._timestamp_newest: datetime = self._TIMESTAMP_MIN
        self._timestamp_oldest: datetime = self._TIMESTAMP_MAX
        self._full_time_range: timedelta = len(self._buffer) * self._sampling_period

    @property
    def sampling_period(self) -> timedelta:
        """Return the sampling period of the ring buffer.

        Returns:
            Sampling period of the ring buffer.
        """
        return self._sampling_period

    @property
    def gaps(self) -> list[Gap]:
        """Get the list of ranges for which no values were provided.

        See definition of dataclass @Gaps for more info.

        Returns:
            List of gaps.
        """
        return self._gaps

    def has_value(self, sample: Sample[QuantityT]) -> bool:
        """Check if a sample has a value and it's not NaN.

        Args:
            sample: sample to check.

        Returns:
            True if the sample has a value and it's not NaN.
        """
        return not (sample.value is None or sample.value.isnan())

    @property
    def maxlen(self) -> int:
        """Get the max length.

        Returns:
            The max amount of items this container can hold.
        """
        return len(self._buffer)

    def update(self, sample: Sample[QuantityT]) -> None:
        """Update the buffer with a new value for the given timestamp.

        Missing values are written as NaN. Be advised that when
        `update()` is called with samples newer than the current time
        + `sampling_period` (as is the case for loading historical data
        after an app restart), a gap of missing data exists. This gap
        does not contain NaN values but simply the old invalid values.
        The list of gaps returned by `gaps()` will reflect this and
        should be used as the only source of truth for unwritten data.

        Args:
            sample: Sample to add to the ringbuffer

        Raises:
            IndexError: When the timestamp to be added is too old.
        """
        # adjust timestamp to be exactly on the sample period time point
        timestamp = self.normalize_timestamp(sample.timestamp)

        # Don't add outdated entries
        if (
            timestamp < self._timestamp_oldest
            and self._timestamp_oldest != self._TIMESTAMP_MAX
        ):
            raise IndexError(
                f"Timestamp {timestamp} too old (cut-off is at {self._timestamp_oldest})."
            )

        # Update timestamps
        prev_newest = self._timestamp_newest
        self._timestamp_newest = max(self._timestamp_newest, timestamp)
        self._timestamp_oldest = self._timestamp_newest - (
            self._full_time_range - self._sampling_period
        )

        # Update data
        if self.has_value(sample):
            assert sample.value is not None
            value = sample.value.base_value
        else:
            value = np.nan
        self._buffer[self.to_internal_index(timestamp)] = value

        self._update_gaps(timestamp, prev_newest, not self.has_value(sample))

    @property
    def time_bound_oldest(self) -> datetime:
        """
        Return the time bounds of the ring buffer.

        Returns:
            The timestamp of the oldest sample of the ring buffer.
        """
        return self._timestamp_oldest

    @property
    def time_bound_newest(self) -> datetime:
        """
        Return the time bounds of the ring buffer.

        Returns:
            The timestamp of the newest sample of the ring buffer
            or None if the buffer is empty.
        """
        return self._timestamp_newest

    @property
    def oldest_timestamp(self) -> datetime | None:
        """Return the oldest timestamp in the buffer.

        Returns:
            The oldest timestamp in the buffer
            or None if the buffer is empty.
        """
        if self.count_valid() == 0:
            return None

        if self.is_missing(self.time_bound_oldest):
            return min(g.end for g in self.gaps)

        return self.time_bound_oldest

    @property
    def newest_timestamp(self) -> datetime | None:
        """Return the newest timestamp in the buffer.

        Returns:
            The newest timestamp in the buffer.
        """
        if self.count_valid() == 0:
            return None

        return self.time_bound_newest

    def to_internal_index(
        self, timestamp: datetime, allow_outside_range: bool = False
    ) -> int:
        """Convert the given timestamp to the position in the buffer.

        !!! Note: This method is meant for advanced use cases and should not generally be used.

        Args:
            timestamp: Timestamp to convert.
            allow_outside_range: If True, don't throw an exception when the
                timestamp is outside our bounds

        Raises:
            IndexError: When requesting a timestamp outside the range this container holds.

        Returns:
            Index where the value for the given timestamp can be found.
        """
        timestamp = self.normalize_timestamp(timestamp)

        if not allow_outside_range and (
            self._timestamp_newest + self._sampling_period < timestamp
            or timestamp < self._timestamp_oldest
        ):
            raise IndexError(
                f"Requested timestamp {timestamp} is "
                f"outside the range [{self._timestamp_oldest} - {self._timestamp_newest}]"
            )

        return self.wrap(
            round(
                (timestamp - self._time_index_alignment).total_seconds()
                / self._sampling_period.total_seconds()
            )
        )

    def get_timestamp(self, index: int) -> datetime | None:
        """Convert the given index to the underlying timestamp.

        Index 0 corresponds to the oldest timestamp in the buffer.
        If negative indices are used, the newest timestamp is used as reference.

        !!!warning

            The resulting timestamp can be outside the range of the buffer.

        Args:
            index: Index to convert.

        Returns:
            Timestamp where the value for the given index can be found.
                Or None if the buffer is empty.
        """
        if self.oldest_timestamp is None:
            return None
        assert self.newest_timestamp is not None
        ref_ts = (
            self.oldest_timestamp
            if index >= 0
            else self.newest_timestamp + self._sampling_period
        )
        return ref_ts + index * self._sampling_period

    def _to_covered_indices(
        self, start: int | None, end: int | None = None
    ) -> tuple[int, int]:
        """Project the given indices via slice onto the covered range.

        Args:
            start: Start index.
            end: End index. Optional, defaults to None.

        Returns:
            tuple of start and end indices on the range currently covered by the buffer.
        """
        return slice(start, end).indices(self.count_covered())[:2]

    def window(
        self,
        start: datetime | int | None,
        end: datetime | int | None,
        *,
        force_copy: bool = True,
        fill_value: float | None = np.nan,
    ) -> FloatArray:
        """Request a copy or view on the data between start timestamp and end timestamp.

        Always request a copy if you keep the data around for longer.
        Otherwise, if the data is not used immediately it could be overwritten.

        Will return a copy in the following cases:
        * The requested time period is crossing the start/end of the buffer.
        * The force_copy parameter was set to True (default True).

        The first case can be avoided by using the appropriate
        `align_to` value in the constructor so that the data lines up
        with the start/end of the buffer.

        This means, if the caller needs to modify the data to account for
        missing entries, they can safely do so.

        Args:
            start: start timestamp of the window.
            end: end timestamp of the window.
            force_copy: optional, default True. If True, will always create a
                copy of the data.
            fill_value: If not None, will use this value to fill missing values.
                If missing values should be filled, force_copy must be True.
                Defaults to NaN to avoid returning outdated data unexpectedly.

        Raises:
            IndexError: When start and end are not both datetime or index.
            ValueError: When fill_value is not None and force_copy is False.

        Returns:
            The requested window
        """
        # We don't want to modify the original buffer
        if fill_value is not None and not force_copy:
            raise ValueError("fill_value only supported for force_copy=True")

        if self.count_covered() == 0:
            return np.array([]) if isinstance(self._buffer, np.ndarray) else []

        # If both are indices or None convert to datetime
        if not isinstance(start, datetime) and not isinstance(end, datetime):
            start, end = self._to_covered_indices(start, end)
            start = self.get_timestamp(start)
            end = self.get_timestamp(end)

        # Here we should have both as datetime
        if not isinstance(start, datetime) or not isinstance(end, datetime):
            raise IndexError(
                f"start ({start}) and end ({end}) must both be either datetime or index."
            )

        # Ensure that the window is within the bounds of the buffer
        assert self.oldest_timestamp is not None and self.newest_timestamp is not None
        start = max(start, self.oldest_timestamp)
        end = min(end, self.newest_timestamp + self._sampling_period)

        if start >= end:
            return np.array([]) if isinstance(self._buffer, np.ndarray) else []

        start_pos = self.to_internal_index(start)
        end_pos = self.to_internal_index(end)

        window = self._wrapped_buffer_window(
            self._buffer, start_pos, end_pos, force_copy
        )

        if fill_value is not None:
            window = self._fill_gaps(window, fill_value, start, self.gaps)
        return window

    def _fill_gaps(
        self,
        data: FloatArray,
        fill_value: float,
        oldest_timestamp: datetime,
        gaps: list[Gap],
    ) -> FloatArray:
        """Fill the gaps in the data with the given fill_value.

        Args:
            data: The data to fill.
            fill_value: The value to fill the gaps with.
            oldest_timestamp: The oldest timestamp in the data.
            gaps: List of gaps to fill.

        Returns:
            The filled data.
        """
        assert isinstance(
            data, (np.ndarray, list)
        ), f"Unsupported data type {type(data)}"
        for gap in gaps:
            start_index = (gap.start - oldest_timestamp) // self._sampling_period
            end_index = (gap.end - oldest_timestamp) // self._sampling_period
            start_index = max(start_index, 0)
            end_index = min(end_index, len(data))
            if start_index < end_index:
                if isinstance(data, np.ndarray):
                    data[start_index:end_index] = fill_value
                elif isinstance(data, list):
                    data[start_index:end_index] = [fill_value] * (
                        end_index - start_index
                    )
        return data

    @staticmethod
    def _wrapped_buffer_window(
        buffer: FloatArray,
        start_pos: int,
        end_pos: int,
        force_copy: bool = True,
    ) -> FloatArray:
        """Get a wrapped window from the given buffer.

        If start_pos == end_pos, the full wrapped buffer is returned starting at start_pos.

        Copies can only be avoided for numpy arrays and when the window is not wrapped.
        Lists of floats are always copies.

        Args:
            buffer: The buffer to get the window from.
            start_pos: The start position of the window in the buffer.
            end_pos: The end position of the window in the buffer (exclusive).
            force_copy: If True, will always create a copy of the data.

        Returns:
            The requested window.
        """
        # Requested window wraps around the ends
        if start_pos >= end_pos:
            if isinstance(buffer, list):
                return buffer[start_pos:] + buffer[0:end_pos]
            assert isinstance(
                buffer, np.ndarray
            ), f"Unsupported buffer type: {type(buffer)}"
            if end_pos > 0:
                return np.concatenate((buffer[start_pos:], buffer[0:end_pos]))
            arr = buffer[start_pos:]
        else:
            arr = buffer[start_pos:end_pos]

        if force_copy:
            return deepcopy(arr)
        return arr

    def is_missing(self, timestamp: datetime) -> bool:
        """Check if the given timestamp falls within a gap.

        Args:
            timestamp: The timestamp to check for missing data.

        Returns:
            True if the given timestamp falls within a gap, False otherwise.
        """
        return any(map(lambda gap: gap.contains(timestamp), self._gaps))

    def _update_gaps(
        self, timestamp: datetime, newest: datetime, record_as_missing: bool
    ) -> None:
        """Update gap list with new timestamp.

        Args:
            timestamp: Timestamp of the new value.
            newest: Timestamp of the newest value before the current update.
            record_as_missing: if `True`, the given timestamp will be recorded as missing.
        """
        found_in_gaps = self.is_missing(timestamp)

        if not record_as_missing:
            # Replace all gaps with one if we went far into then future
            if self._timestamp_newest - newest >= self._full_time_range:
                self._gaps = [
                    Gap(start=self._timestamp_oldest, end=self._timestamp_newest)
                ]
                return

            # Check if we created a gap with the addition of the new value
            if not found_in_gaps and timestamp > newest + self._sampling_period:
                self._gaps.append(
                    Gap(start=newest + self._sampling_period, end=timestamp)
                )

        # New missing entry that is not already in a gap?
        if record_as_missing:
            if not found_in_gaps:
                # If there are no gaps and the new value is not subsequent to the
                # newest value, we need to start the new gap after the newest value
                start_gap = min(newest + self._sampling_period, timestamp)
                self._gaps.append(
                    Gap(start=start_gap, end=timestamp + self._sampling_period)
                )
        elif len(self._gaps) > 0:
            if found_in_gaps:
                self._remove_gap(timestamp)

        self._cleanup_gaps()

    def _cleanup_gaps(self) -> None:
        """Clean up the list of gaps.

        * Merge existing gaps
        * remove overlaps
        * delete outdated gaps
        """
        self._gaps = sorted(self._gaps, key=lambda x: x.start.timestamp())

        i = 0
        while i < len(self._gaps):
            w_1 = self._gaps[i]
            if i < len(self._gaps) - 1:
                w_2 = self._gaps[i + 1]
            else:
                w_2 = None

            # Delete out-of-date gaps
            if w_1.end <= self._timestamp_oldest:
                del self._gaps[i]
            # Update start of gap if it's rolled out of the buffer
            elif w_1.start < self._timestamp_oldest:
                self._gaps[i].start = self._timestamp_oldest
            # If w2 is a subset of w1 we can delete it
            elif w_2 and w_1.start <= w_2.start and w_1.end >= w_2.end:
                del self._gaps[i + 1]
            # If the gaps are direct neighbors, merge them
            elif w_2 and w_1.end >= w_2.start:
                w_1.end = w_2.end
                del self._gaps[i + 1]
            else:
                i += 1

    def _remove_gap(self, timestamp: datetime) -> None:
        """Update the list of gaps to not contain the given timestamp.

        Args:
            timestamp: Timestamp that is no longer missing.
        """
        gap_index, gap = next(
            filter(
                lambda gap: gap[1].contains(timestamp),
                enumerate(self._gaps),
            ),
            (0, None),
        )

        if gap is None:
            return

        if gap.start == timestamp:
            # Is the whole gap consisting only of the timestamp?
            if gap.end == timestamp + self._sampling_period:
                del self._gaps[gap_index]
            # Otherwise, make the gap smaller
            else:
                gap.start = timestamp + self._sampling_period
        # Is the timestamp at the end? Shrink by one then
        elif gap.end - self._sampling_period == timestamp:
            gap.end = timestamp
        # Otherwise it must be in the middle and we need to create a new
        # gap
        else:
            new_gap = deepcopy(gap)
            gap.end = timestamp
            new_gap.start = timestamp + self._sampling_period
            self._gaps.append(new_gap)

    def normalize_timestamp(self, timestamp: datetime) -> datetime:
        """Normalize the given timestamp to fall exactly on the resampling period.

        Args:
            timestamp: The timestamp to normalize.

        Returns:
            The normalized timestamp.
        """
        num_samples, remainder = divmod(
            (timestamp - self._time_index_alignment), self._sampling_period
        )

        # Round towards the closer number and towards the even one in case of
        # equal distance
        if remainder != timedelta(0) and (
            self._sampling_period / 2 == remainder
            and num_samples % 2 != 0
            or self._sampling_period / 2 < remainder
        ):
            num_samples += 1

        normalized_timestamp = (
            self._time_index_alignment + num_samples * self._sampling_period
        )

        return normalized_timestamp

    def wrap(self, index: int) -> int:
        """Normalize the given index to fit in the buffer by wrapping it around.

        Args:
            index: index to normalize.

        Returns:
            An index that will be within maxlen.
        """
        return index % self.maxlen

    @overload
    def __getitem__(self, index_or_slice: SupportsIndex) -> float:
        """Get item at requested position.

        No wrapping of the index will be done.
        Create a feature request if you require this function.

        Args:
            index_or_slice: Index of the requested data.

        Returns:
            The requested value.
        """

    @overload
    def __getitem__(self, index_or_slice: slice) -> FloatArray:
        """Get the data described by the given slice.

        No wrapping of the index will be done.
        Create a feature request if you require this function.

        Args:
            index_or_slice: Slice specification of where the requested data is.

        Returns:
            The requested slice.
        """

    def __getitem__(self, index_or_slice: SupportsIndex | slice) -> float | FloatArray:
        """Get item or slice at requested position.

        No wrapping of the index will be done.
        Create a feature request if you require this function.

        Args:
            index_or_slice: Index or slice specification of the requested data.

        Returns:
            The requested value or slice.
        """
        return self._buffer.__getitem__(index_or_slice)

    def _covered_time_range(self) -> timedelta:
        """Return the time range that is covered by the oldest and newest valid samples.

        Returns:
            The time range between the oldest and newest valid samples or 0 if
                there are is no time range covered.
        """
        if not self.oldest_timestamp:
            return timedelta(0)

        assert (
            self.newest_timestamp is not None
        ), "Newest timestamp cannot be None here."
        return self.newest_timestamp - self.oldest_timestamp + self._sampling_period

    def count_covered(self) -> int:
        """Count the number of samples that are covered by the oldest and newest valid samples.

        Returns:
            The count of samples between the oldest and newest (inclusive) valid samples
                or 0 if there are is no time range covered.
        """
        return int(
            self._covered_time_range().total_seconds()
            // self._sampling_period.total_seconds()
        )

    def count_valid(self) -> int:
        """Count the number of valid items that this buffer currently holds.

        Returns:
            The number of valid items in this buffer.
        """
        if self._timestamp_newest == self._TIMESTAMP_MIN:
            return 0

        # Sum of all elements in the gap ranges
        sum_missing_entries = max(
            0,
            sum(
                (
                    gap.end
                    # Don't look further back than oldest timestamp
                    - max(gap.start, self._timestamp_oldest)
                )
                // self._sampling_period
                for gap in self._gaps
            ),
        )

        start_pos = self.to_internal_index(self._timestamp_oldest)
        end_pos = self.to_internal_index(self._timestamp_newest)

        if end_pos < start_pos:
            return len(self._buffer) - start_pos + end_pos + 1 - sum_missing_entries

        return end_pos + 1 - start_pos - sum_missing_entries



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A builder for creating formula engines that operate on resampled component metrics."""

from __future__ import annotations

from collections.abc import Callable

from frequenz.channels import Receiver, Sender
from frequenz.client.microgrid import ComponentMetricId
from frequenz.quantities import Quantity

from ..._internal._channels import ChannelRegistry
from ...microgrid._data_sourcing import ComponentMetricRequest
from .._base_types import QuantityT, Sample
from ._formula_engine import FormulaBuilder, FormulaEngine
from ._formula_steps import FallbackMetricFetcher
from ._tokenizer import Tokenizer, TokenType


class ResampledFormulaBuilder(FormulaBuilder[QuantityT]):
    """Provides a way to build a FormulaEngine from resampled data streams."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        namespace: str,
        formula_name: str,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
        metric_id: ComponentMetricId,
        create_method: Callable[[float], QuantityT],
    ) -> None:
        """Create a `ResampledFormulaBuilder` instance.

        Args:
            namespace: The unique namespace to allow reuse of streams in the data
                pipeline.
            formula_name: A name for the formula.
            channel_registry: The channel registry instance shared with the resampling
                and the data sourcing actors.
            resampler_subscription_sender: A sender to send metric requests to the
                resampling actor.
            metric_id: A metric ID to fetch for all components in this formula.
            create_method: A method to generate the output `Sample` value with.  If the
                formula is for generating power values, this would be
                `Power.from_watts`, for example.
        """
        self._channel_registry: ChannelRegistry = channel_registry
        self._resampler_subscription_sender: Sender[ComponentMetricRequest] = (
            resampler_subscription_sender
        )
        self._namespace: str = namespace
        self._metric_id: ComponentMetricId = metric_id
        self._resampler_requests: list[ComponentMetricRequest] = []
        super().__init__(formula_name, create_method)

    def _get_resampled_receiver(
        self, component_id: int, metric_id: ComponentMetricId
    ) -> Receiver[Sample[QuantityT]]:
        """Get a receiver with the resampled data for the given component id.

        Args:
            component_id: The component id for which to get a resampled data receiver.
            metric_id: A metric ID to fetch for all components in this formula.

        Returns:
            A receiver to stream resampled data for the given component id.
        """
        request = ComponentMetricRequest(self._namespace, component_id, metric_id, None)
        self._resampler_requests.append(request)
        resampled_channel = self._channel_registry.get_or_create(
            Sample[Quantity], request.get_channel_name()
        )
        resampled_receiver = resampled_channel.new_receiver().map(
            lambda sample: Sample(
                sample.timestamp,
                (
                    self._create_method(sample.value.base_value)
                    if sample.value is not None
                    else None
                ),
            )
        )
        return resampled_receiver

    async def subscribe(self) -> None:
        """Subscribe to all resampled component metric streams."""
        for request in self._resampler_requests:
            await self._resampler_subscription_sender.send(request)

    def push_component_metric(
        self,
        component_id: int,
        *,
        nones_are_zeros: bool,
        fallback: FallbackMetricFetcher[QuantityT] | None = None,
    ) -> None:
        """Push a resampled component metric stream to the formula engine.

        Args:
            component_id: The component id for which to push a metric fetcher.
            nones_are_zeros: Whether to treat None values from the stream as 0s.  If
                False, the returned value will be a None.
            fallback: Metric fetcher to use if primary one start sending
                invalid data (e.g. due to a component stop). If None the data from
                primary metric fetcher will be returned.
        """
        receiver = self._get_resampled_receiver(component_id, self._metric_id)
        self.push_metric(
            f"#{component_id}",
            receiver,
            nones_are_zeros=nones_are_zeros,
            fallback=fallback,
        )

    def from_string(
        self,
        formula: str,
        *,
        nones_are_zeros: bool,
    ) -> FormulaEngine[QuantityT]:
        """Construct a `FormulaEngine` from the given formula string.

        Formulas can have Component IDs that are preceeded by a pound symbol("#"), and
        these operators: +, -, *, /, (, ).

        For example, the input string: "#20 + #5" is a formula for adding metrics from
        two components with ids 20 and 5.

        Args:
            formula: A string formula.
            nones_are_zeros: Whether to treat None values from the stream as 0s.  If
                False, the returned value will be a None.

        Returns:
            A FormulaEngine instance corresponding to the given formula.

        Raises:
            ValueError: when there is an unknown token type.
        """
        tokenizer = Tokenizer(formula)

        for token in tokenizer:
            if token.type == TokenType.COMPONENT_METRIC:
                self.push_component_metric(
                    int(token.value), nones_are_zeros=nones_are_zeros
                )
            elif token.type == TokenType.OPER:
                self.push_oper(token.value)
            else:
                raise ValueError(f"Unknown token type: {token}")

        return self.build()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Background service implementation."""

import abc
import asyncio
import collections.abc
from types import TracebackType
from typing import Any, Self


class BackgroundService(abc.ABC):
    """A background service that can be started and stopped.

    A background service is a service that runs in the background spawning one or more
    tasks. The service can be [started][frequenz.sdk.actor.BackgroundService.start]
    and [stopped][frequenz.sdk.actor.BackgroundService.stop] and can work as an
    async context manager to provide deterministic cleanup.

    To implement a background service, subclasses must implement the
    [`start()`][frequenz.sdk.actor.BackgroundService.start] method, which should
    start the background tasks needed by the service, and add them to the `_tasks`
    protected attribute.

    If you need to collect results or handle exceptions of the tasks when stopping the
    service, then you need to also override the
    [`stop()`][frequenz.sdk.actor.BackgroundService.stop] method, as the base
    implementation does not collect any results and re-raises all exceptions.

    !!! warning

        As background services manage [`asyncio.Task`][] objects, a reference to them
        must be held for as long as the background service is expected to be running,
        otherwise its tasks will be cancelled and the service will stop. For more
        information, please refer to the [Python `asyncio`
        documentation](https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task).

    Example:
        ```python
        import datetime
        import asyncio

        class Clock(BackgroundService):
            def __init__(self, resolution_s: float, *, name: str | None = None) -> None:
                super().__init__(name=name)
                self._resolution_s = resolution_s

            def start(self) -> None:
                self._tasks.add(asyncio.create_task(self._tick()))

            async def _tick(self) -> None:
                while True:
                    await asyncio.sleep(self._resolution_s)
                    print(datetime.datetime.now())

        async def main() -> None:
            # As an async context manager
            async with Clock(resolution_s=1):
                await asyncio.sleep(5)

            # Manual start/stop (only use if necessary, as cleanup is more complicated)
            clock = Clock(resolution_s=1)
            clock.start()
            await asyncio.sleep(5)
            await clock.stop()

        asyncio.run(main())
        ```
    """

    def __init__(self, *, name: str | None = None) -> None:
        """Initialize this BackgroundService.

        Args:
            name: The name of this background service. If `None`, `str(id(self))` will
                be used. This is used mostly for debugging purposes.
        """
        self._name: str = str(id(self)) if name is None else name
        self._tasks: set[asyncio.Task[Any]] = set()

    @abc.abstractmethod
    def start(self) -> None:
        """Start this background service."""

    @property
    def name(self) -> str:
        """The name of this background service.

        Returns:
            The name of this background service.
        """
        return self._name

    @property
    def tasks(self) -> collections.abc.Set[asyncio.Task[Any]]:
        """Return the set of running tasks spawned by this background service.

        Users typically should not modify the tasks in the returned set and only use
        them for informational purposes.

        !!! danger

            Changing the returned tasks may lead to unexpected behavior, don't do it
            unless the class explicitly documents it is safe to do so.

        Returns:
            The set of running tasks spawned by this background service.
        """
        return self._tasks

    @property
    def is_running(self) -> bool:
        """Return whether this background service is running.

        A service is considered running when at least one task is running.

        Returns:
            Whether this background service is running.
        """
        return any(not task.done() for task in self._tasks)

    def cancel(self, msg: str | None = None) -> None:
        """Cancel all running tasks spawned by this background service.

        Args:
            msg: The message to be passed to the tasks being cancelled.
        """
        for task in self._tasks:
            task.cancel(msg)

    # We need the noqa because pydoclint can't figure out `rest` is
    # a `BaseExceptionGroup` instance.
    async def stop(self, msg: str | None = None) -> None:  # noqa: DOC503
        """Stop this background service.

        This method cancels all running tasks spawned by this service and waits for them
        to finish.

        Args:
            msg: The message to be passed to the tasks being cancelled.

        Raises:
            BaseExceptionGroup: If any of the tasks spawned by this service raised an
                exception.
        """
        if not self._tasks:
            return
        self.cancel(msg)
        try:
            await self.wait()
        except BaseExceptionGroup as exc_group:
            # We want to ignore CancelledError here as we explicitly cancelled all the
            # tasks.
            _, rest = exc_group.split(asyncio.CancelledError)
            if rest is not None:
                # We are filtering out from an exception group, we really don't want to
                # add the exceptions we just filtered by adding a from clause here.
                raise rest  # pylint: disable=raise-missing-from

    async def __aenter__(self) -> Self:
        """Enter an async context.

        Start this background service.

        Returns:
            This background service.
        """
        self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit an async context.

        Stop this background service.

        Args:
            exc_type: The type of the exception raised, if any.
            exc_val: The exception raised, if any.
            exc_tb: The traceback of the exception raised, if any.
        """
        await self.stop()

    async def wait(self) -> None:
        """Wait this background service to finish.

        Wait until all background service tasks are finished.

        Raises:
            BaseExceptionGroup: If any of the tasks spawned by this service raised an
                exception (`CancelError` is not considered an error and not returned in
                the exception group).
        """
        # We need to account for tasks that were created between when we started
        # awaiting and we finished awaiting.
        while self._tasks:
            done, pending = await asyncio.wait(self._tasks)
            assert not pending

            # We remove the done tasks, but there might be new ones created after we
            # started waiting.
            self._tasks = self._tasks - done

            exceptions: list[BaseException] = []
            for task in done:
                try:
                    # This will raise a CancelledError if the task was cancelled or any
                    # other exception if the task raised one.
                    _ = task.result()
                except BaseException as error:  # pylint: disable=broad-except
                    exceptions.append(error)
            if exceptions:
                raise BaseExceptionGroup(
                    f"Error while stopping background service {self}", exceptions
                )

    def __await__(self) -> collections.abc.Generator[None, None, None]:
        """Await this background service.

        An awaited background service will wait for all its tasks to finish.

        Returns:
            An implementation-specific generator for the awaitable.
        """
        return self.wait().__await__()

    def __del__(self) -> None:
        """Destroy this instance.

        Cancel all running tasks spawned by this background service.
        """
        self.cancel("{self!r} was deleted")

    def __repr__(self) -> str:
        """Return a string representation of this instance.

        Returns:
            A string representation of this instance.
        """
        return f"{type(self).__name__}(name={self._name!r}, tasks={self._tasks!r})"

    def __str__(self) -> str:
        """Return a string representation of this instance.

        Returns:
            A string representation of this instance.
        """
        return f"{type(self).__name__}[{self._name}]"



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH
"""Definition of the user request."""


import dataclasses
from collections import abc

from frequenz.quantities import Power


@dataclasses.dataclass
class Request:
    """Request to set power to the `PowerDistributingActor`."""

    power: Power
    """The requested power."""

    component_ids: abc.Set[int]
    """The component ids of the components to be used for this request."""

    adjust_power: bool = True
    """Whether to adjust the power to match the bounds.

    If `True`, the power will be adjusted (lowered) to match the bounds, so
    only the reduced power will be set.

    If `False` and the power is outside the available bounds, the request will
    fail and be replied to with an `OutOfBound` result.
    """



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for Grid Power."""


from frequenz.client.microgrid import Component, ComponentMetricId
from frequenz.quantities import Power

from .._formula_engine import FormulaEngine
from ._fallback_formula_metric_fetcher import FallbackFormulaMetricFetcher
from ._formula_generator import FormulaGeneratorConfig
from ._grid_power_formula_base import GridPowerFormulaBase
from ._simple_formula import SimplePowerFormula


class GridPowerFormula(GridPowerFormulaBase[Power]):
    """Creates a formula engine from the component graph for calculating grid power."""

    def generate(  # noqa: DOC502
        # * ComponentNotFound is raised indirectly by _get_grid_component_successors
        self,
    ) -> FormulaEngine[Power]:
        """Generate a formula for calculating grid power from the component graph.

        Returns:
            A formula engine that will calculate grid power values.

        Raises:
            ComponentNotFound: when the component graph doesn't have a `GRID` component.
        """
        builder = self._get_builder(
            "grid-power",
            ComponentMetricId.ACTIVE_POWER,
            Power.from_watts,
        )
        return self._generate(builder)

    def _get_fallback_formulas(
        self, components: set[Component]
    ) -> dict[Component, FallbackFormulaMetricFetcher[Power] | None]:
        """Find primary and fallback components and create fallback formulas.

        The primary component is the one that will be used to calculate the grid power.
        If it is not available, the fallback formula will be used instead.
        Fallback formulas calculate the grid power using the fallback components.
        Fallback formulas are wrapped in `FallbackFormulaMetricFetcher`.

        Args:
            components: The producer components.

        Returns:
            A dictionary mapping primary components to their FallbackFormulaMetricFetcher.
        """
        fallbacks = self._get_metric_fallback_components(components)

        fallback_formulas: dict[
            Component, FallbackFormulaMetricFetcher[Power] | None
        ] = {}

        for primary_component, fallback_components in fallbacks.items():
            if len(fallback_components) == 0:
                fallback_formulas[primary_component] = None
                continue

            fallback_ids = [c.component_id for c in fallback_components]
            generator = SimplePowerFormula(
                f"{self._namespace}_fallback_{fallback_ids}",
                self._channel_registry,
                self._resampler_subscription_sender,
                FormulaGeneratorConfig(
                    component_ids=set(fallback_ids),
                    allow_fallback=False,
                ),
            )

            fallback_formulas[primary_component] = FallbackFormulaMetricFetcher(
                generator
            )

        return fallback_formulas



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""A power manager implementation."""

from ._base_classes import Algorithm, Proposal, ReportRequest, _Report
from ._power_managing_actor import PowerManagingActor

__all__ = [
    "Algorithm",
    "PowerManagingActor",
    "Proposal",
    "_Report",
    "ReportRequest",
]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Read and update config variables."""

from ._config_managing import ConfigManagingActor
from ._logging_config_updater import LoggerConfig, LoggingConfig, LoggingConfigUpdater
from ._util import load_config

__all__ = [
    "ConfigManagingActor",
    "LoggingConfig",
    "LoggerConfig",
    "LoggingConfigUpdater",
    "load_config",
]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for 3-phase Grid Power."""

from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Power

from .._formula_engine import FormulaEngine, FormulaEngine3Phase
from ._formula_generator import FormulaGenerator


class GridPower3PhaseFormula(FormulaGenerator[Power]):
    """Create a formula engine for calculating the grid 3-phase power."""

    def generate(  # noqa: DOC502
        # ComponentNotFound is raised indirectly by _get_grid_component_successors
        self,
    ) -> FormulaEngine3Phase[Power]:
        """Generate a formula for calculating grid 3-phase power.

        Raises:
            ComponentNotFound: when the component graph doesn't have a `GRID` component.

        Returns:
            A formula engine that will calculate grid 3-phase power values.
        """
        grid_successors = self._get_grid_component_successors()

        return FormulaEngine3Phase(
            "grid-power-3-phase",
            Power.from_watts,
            (
                self._gen_phase_formula(
                    grid_successors, ComponentMetricId.ACTIVE_POWER_PHASE_1
                ),
                self._gen_phase_formula(
                    grid_successors, ComponentMetricId.ACTIVE_POWER_PHASE_2
                ),
                self._gen_phase_formula(
                    grid_successors, ComponentMetricId.ACTIVE_POWER_PHASE_3
                ),
            ),
        )

    def _gen_phase_formula(
        self,
        grid_successors: set[Component],
        metric_id: ComponentMetricId,
    ) -> FormulaEngine[Power]:
        """Generate a formula for calculating grid 3-phase power from the component graph.

        Generate a formula that adds values from all components that are directly
        connected to the grid.

        Args:
            grid_successors: The set of components that are directly connected to the grid.
            metric_id: The metric to use for the formula.

        Returns:
            A formula engine that will calculate grid 3-phase power values.
        """
        formula_builder = self._get_builder(
            "grid-power-3-phase", metric_id, Power.from_watts
        )

        for idx, comp in enumerate(grid_successors):
            # When inverters or EV chargers produce `None` samples, they are
            # excluded from the calculation by treating their `None` values
            # as `0`s.
            #
            # This is not possible for Meters, so when they produce `None`
            # values, those values get propagated as the output.
            if comp.category in (
                ComponentCategory.INVERTER,
                ComponentCategory.EV_CHARGER,
            ):
                nones_are_zeros = True
            elif comp.category == ComponentCategory.METER:
                nones_are_zeros = False
            else:
                continue

            if idx > 0:
                formula_builder.push_oper("+")

            formula_builder.push_component_metric(
                comp.component_id, nones_are_zeros=nones_are_zeros
            )

        return formula_builder.build()



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Tracking of the blocking status of a component."""

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone


@dataclass(kw_only=True)
class BlockingStatus:
    """Tracking of the blocking status of a component."""

    min_duration: timedelta
    """The minimum blocking duration."""

    max_duration: timedelta
    """The maximum blocking duration."""

    last_blocking_duration: timedelta = timedelta(seconds=0.0)
    """Last blocking duration."""

    blocked_until: datetime | None = None
    """Time until which the component is blocked."""

    def __post_init__(self) -> None:
        """Validate the blocking duration."""
        assert self.min_duration <= self.max_duration, (
            f"Minimum blocking duration ({self.min_duration}) cannot be greater "
            f"than maximum blocking duration ({self.max_duration})"
        )
        self.last_blocking_duration = self.min_duration
        self._timedelta_zero = timedelta(seconds=0.0)

    def block(self) -> timedelta:
        """Set the component as blocked.

        Returns:
            The duration for which the component is blocked.
        """
        now = datetime.now(tz=timezone.utc)

        # If is not blocked
        if self.blocked_until is None:
            self.last_blocking_duration = self.min_duration
            self.blocked_until = now + self.last_blocking_duration
            return self.last_blocking_duration

        # If still blocked, then do nothing
        if self.blocked_until > now:
            return self._timedelta_zero

        # If previous blocking time expired, then blocked it once again.
        # Increase last blocking time, unless it reach the maximum.
        self.last_blocking_duration = min(
            2 * self.last_blocking_duration, self.max_duration
        )
        self.blocked_until = now + self.last_blocking_duration

        return self.last_blocking_duration

    def unblock(self) -> None:
        """Set the component as unblocked."""
        self.blocked_until = None

    def is_blocked(self) -> bool:
        """Check if the component is blocked.

        Returns:
            True if battery is blocked, False otherwise.
        """
        if self.blocked_until is None:
            return False
        return self.blocked_until > datetime.now(tz=timezone.utc)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Fuse data class."""
from __future__ import annotations

from dataclasses import dataclass

from frequenz.quantities import Current


@dataclass(frozen=True)
class Fuse:
    """Fuse data class."""

    max_current: Current
    """Rated current of the fuse."""



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

# pylint: disable=too-many-lines

"""A formula engine that can apply formulas on streaming data."""

from __future__ import annotations

import asyncio
import logging
from abc import ABC
from collections import deque
from collections.abc import Callable
from typing import Any, Generic, Self, TypeVar

from frequenz.channels import Broadcast, Receiver
from frequenz.quantities import Quantity

from ..._internal._asyncio import cancel_and_await
from .._base_types import QuantityT, Sample, Sample3Phase
from ._formula_evaluator import FormulaEvaluator
from ._formula_formatter import format_formula
from ._formula_steps import (
    Adder,
    Clipper,
    ConstantValue,
    Consumption,
    Divider,
    FallbackMetricFetcher,
    FormulaStep,
    Maximizer,
    MetricFetcher,
    Minimizer,
    Multiplier,
    OpenParen,
    Production,
    Subtractor,
)
from ._tokenizer import TokenType

_logger = logging.Logger(__name__)

_operator_precedence = {
    "max": 0,
    "min": 1,
    "consumption": 2,
    "production": 3,
    "(": 4,
    "/": 5,
    "*": 6,
    "-": 7,
    "+": 8,
    ")": 9,
}
"""The dictionary of operator precedence for the shunting yard algorithm."""


class FormulaEngine(Generic[QuantityT]):
    """An engine to apply formulas on resampled data streams.

    Please refer to the [module documentation][frequenz.sdk.timeseries.formula_engine]
    for more information on how formula engines are used throughout the SDK.

    Example: Streaming the power of a battery pool.
        ```python
        from frequenz.sdk import microgrid

        battery_pool = microgrid.new_battery_pool(priority=5)

        async for power in battery_pool.power.new_receiver():
            print(f"{power=}")
        ```

    Example: Composition of formula engines.
        ```python
        from frequenz.sdk import microgrid

        battery_pool = microgrid.new_battery_pool(priority=5)
        ev_charger_pool = microgrid.new_ev_charger_pool(priority=5)
        grid = microgrid.grid()

        # apply operations on formula engines to create a formula engine that would
        # apply these operations on the corresponding data streams.
        net_power = (
            grid.power - (battery_pool.power + ev_charger_pool.power)
        ).build("net_power")

        async for power in net_power.new_receiver():
            print(f"{power=}")
        ```
    """

    def __init__(
        self,
        builder: FormulaBuilder[QuantityT],
        create_method: Callable[[float], QuantityT],
    ) -> None:
        """Create a `FormulaEngine` instance.

        Args:
            builder: A `FormulaBuilder` instance to get the formula steps and metric
                fetchers from.
            create_method: A method to generate the output `Sample` value with.  If the
                formula is for generating power values, this would be
                `Power.from_watts`, for example.
        """
        self._higher_order_builder = HigherOrderFormulaBuilder
        self._name: str = builder.name
        self._builder: FormulaBuilder[QuantityT] = builder
        self._create_method: Callable[[float], QuantityT] = create_method
        self._channel: Broadcast[Sample[QuantityT]] = Broadcast(name=self._name)
        self._task: asyncio.Task[None] | None = None

    async def _stop(self) -> None:
        """Stop a running formula engine."""
        if self._task is None:
            return
        await cancel_and_await(self._task)

    @classmethod
    def from_receiver(
        cls,
        name: str,
        receiver: Receiver[Sample[QuantityT]],
        create_method: Callable[[float], QuantityT],
        *,
        nones_are_zeros: bool = False,
    ) -> FormulaEngine[QuantityT]:
        """
        Create a formula engine from a receiver.

        Can be used to compose a formula engine with a receiver. When composing
        the new engine with other engines, make sure that receiver gets data
        from the same resampler and that the `create_method`s match.

        Example:
            ```python
            from frequenz.sdk import microgrid
            from frequenz.quantities import Power

            async def run() -> None:
                producer_power_engine = microgrid.producer().power
                consumer_power_recv = microgrid.consumer().power.new_receiver()

                excess_power_recv = (
                    (
                        producer_power_engine
                        + FormulaEngine.from_receiver(
                            "consumer power",
                            consumer_power_recv,
                            Power.from_watts,
                        )
                    )
                    .build("excess power")
                    .new_receiver()
                )

            asyncio.run(run())
            ```

        Args:
            name: A name for the formula engine.
            receiver: A receiver that streams `Sample`s.
            create_method: A method to generate the output `Sample` value with,
                e.g. `Power.from_watts`.
            nones_are_zeros: If `True`, `None` values in the receiver are treated as 0.

        Returns:
            A formula engine that streams the `Sample`s from the receiver.
        """
        builder = FormulaBuilder(name, create_method)
        builder.push_metric(name, receiver, nones_are_zeros=nones_are_zeros)
        return cls(builder, create_method)

    def __add__(
        self,
        other: (
            FormulaEngine[QuantityT] | HigherOrderFormulaBuilder[QuantityT] | QuantityT
        ),
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """Return a formula builder that adds (data in) `other` to `self`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder(self, self._create_method) + other

    def __sub__(
        self,
        other: (
            FormulaEngine[QuantityT] | HigherOrderFormulaBuilder[QuantityT] | QuantityT
        ),
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """Return a formula builder that subtracts (data in) `other` from `self`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder(self, self._create_method) - other

    def __mul__(
        self,
        other: FormulaEngine[QuantityT] | HigherOrderFormulaBuilder[QuantityT] | float,
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """Return a formula builder that multiplies (data in) `self` with `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder(self, self._create_method) * other

    def __truediv__(
        self,
        other: FormulaEngine[QuantityT] | HigherOrderFormulaBuilder[QuantityT] | float,
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """Return a formula builder that divides (data in) `self` by `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder(self, self._create_method) / other

    def max(
        self,
        other: (
            FormulaEngine[QuantityT] | HigherOrderFormulaBuilder[QuantityT] | QuantityT
        ),
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """Return a formula engine that outputs the maximum of `self` and `other`.

        Args:
            other: A formula receiver, a formula builder or a QuantityT instance
                corresponding to a sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder(self, self._create_method).max(other)

    def min(
        self,
        other: (
            FormulaEngine[QuantityT] | HigherOrderFormulaBuilder[QuantityT] | QuantityT
        ),
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """Return a formula engine that outputs the minimum of `self` and `other`.

        Args:
            other: A formula receiver, a formula builder or a QuantityT instance
                corresponding to a sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder(self, self._create_method).min(other)

    def consumption(
        self,
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """
        Return a formula builder that applies the consumption operator on `self`.

        The consumption operator returns either the identity if the power value is
        positive or 0.
        """
        return HigherOrderFormulaBuilder(self, self._create_method).consumption()

    def production(
        self,
    ) -> HigherOrderFormulaBuilder[QuantityT]:
        """
        Return a formula builder that applies the production operator on `self`.

        The production operator returns either the absolute value if the power value is
        negative or 0.
        """
        return HigherOrderFormulaBuilder(self, self._create_method).production()

    async def _run(self) -> None:
        await self._builder.subscribe()
        steps, metric_fetchers = self._builder.finalize()
        evaluator = FormulaEvaluator[QuantityT](
            self._name, steps, metric_fetchers, self._create_method
        )
        sender = self._channel.new_sender()
        while True:
            try:
                msg = await evaluator.apply()
            except asyncio.CancelledError:
                _logger.exception("FormulaEngine task cancelled: %s", self._name)
                raise
            except Exception as err:  # pylint: disable=broad-except
                _logger.warning(
                    "Formula application failed: %s. Error: %s", self._name, err
                )
            else:
                await sender.send(msg)

    def __str__(self) -> str:
        """Return a string representation of the formula.

        Returns:
            A string representation of the formula.
        """
        steps = (
            self._builder._build_stack
            if len(self._builder._build_stack) > 0
            else self._builder._steps
        )
        return format_formula(steps)

    def new_receiver(
        self, name: str | None = None, max_size: int = 50
    ) -> Receiver[Sample[QuantityT]]:
        """Create a new receiver that streams the output of the formula engine.

        Args:
            name: An optional name for the receiver.
            max_size: The size of the receiver's buffer.

        Returns:
            A receiver that streams output `Sample`s from the formula engine.
        """
        if self._task is None:
            self._task = asyncio.create_task(self._run())

        recv = self._channel.new_receiver(name=name, limit=max_size)

        # This is a hack to ensure that the lifetime of the engine is tied to the
        # lifetime of the receiver.  This is necessary because the engine is a task that
        # runs forever, and in cases where higher order built for example with the below
        # idiom, the user would hold no references to the engine and it could get
        # garbage collected before the receiver.  This behaviour is explained in the
        # `asyncio.create_task` docs here:
        # https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
        #
        #     formula = (grid_power_engine + bat_power_engine).build().new_receiver()
        recv._engine_reference = self  # type: ignore # pylint: disable=protected-access
        return recv


class FormulaEngine3Phase(Generic[QuantityT]):
    """An engine to apply formulas on 3-phase resampled data streams.

    Please refer to the [module documentation][frequenz.sdk.timeseries.formula_engine]
    for more information on how formula engines are used throughout the SDK.

    Example: Streaming the current of an EV charger pool.
        ```python
        from frequenz.sdk import microgrid

        ev_charger_pool = microgrid.new_ev_charger_pool(priority=5)

        async for sample in ev_charger_pool.current_per_phase.new_receiver():
            print(f"Current: {sample}")
        ```

    Example: Composition of formula engines.
        ```python
        from frequenz.sdk import microgrid

        ev_charger_pool = microgrid.new_ev_charger_pool(priority=5)
        grid = microgrid.grid()

        # Calculate grid consumption current that's not used by the EV chargers
        other_current = (grid.current_per_phase - ev_charger_pool.current_per_phase).build(
            "other_current"
        )

        async for sample in other_current.new_receiver():
            print(f"Other current: {sample}")
        ```
    """

    def __init__(
        self,
        name: str,
        create_method: Callable[[float], QuantityT],
        phase_streams: tuple[
            FormulaEngine[QuantityT],
            FormulaEngine[QuantityT],
            FormulaEngine[QuantityT],
        ],
    ) -> None:
        """Create a `FormulaEngine3Phase` instance.

        Args:
            name: A name for the formula.
            create_method: A method to generate the output `Sample` value with.  If the
                formula is for generating power values, this would be
                `Power.from_watts`, for example.
            phase_streams: output streams of formula engines running per-phase formulas.
        """
        self._higher_order_builder = HigherOrderFormulaBuilder3Phase
        self._name: str = name
        self._create_method: Callable[[float], QuantityT] = create_method
        self._channel: Broadcast[Sample3Phase[QuantityT]] = Broadcast(name=self._name)
        self._task: asyncio.Task[None] | None = None
        self._streams: tuple[
            FormulaEngine[QuantityT],
            FormulaEngine[QuantityT],
            FormulaEngine[QuantityT],
        ] = phase_streams

    async def _stop(self) -> None:
        """Stop a running formula engine."""
        if self._task is None:
            return
        await cancel_and_await(self._task)

    def __add__(
        self,
        other: (
            FormulaEngine3Phase[QuantityT] | HigherOrderFormulaBuilder3Phase[QuantityT]
        ),
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """Return a formula builder that adds (data in) `other` to `self`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method) + other

    def __sub__(
        self: FormulaEngine3Phase[QuantityT],
        other: (
            FormulaEngine3Phase[QuantityT] | HigherOrderFormulaBuilder3Phase[QuantityT]
        ),
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """Return a formula builder that subtracts (data in) `other` from `self`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method) - other

    def __mul__(
        self: FormulaEngine3Phase[QuantityT],
        other: (
            FormulaEngine3Phase[QuantityT] | HigherOrderFormulaBuilder3Phase[QuantityT]
        ),
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """Return a formula builder that multiplies (data in) `self` with `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method) * other

    def __truediv__(
        self: FormulaEngine3Phase[QuantityT],
        other: (
            FormulaEngine3Phase[QuantityT] | HigherOrderFormulaBuilder3Phase[QuantityT]
        ),
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """Return a formula builder that divides (data in) `self` by `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method) / other

    def max(
        self: FormulaEngine3Phase[QuantityT],
        other: (
            FormulaEngine3Phase[QuantityT] | HigherOrderFormulaBuilder3Phase[QuantityT]
        ),
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """Return a formula engine that outputs the maximum of `self` and `other`.

        Args:
            other: A formula receiver, a formula builder or a QuantityT instance
                corresponding to a sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method).max(other)

    def min(
        self: FormulaEngine3Phase[QuantityT],
        other: (
            FormulaEngine3Phase[QuantityT] | HigherOrderFormulaBuilder3Phase[QuantityT]
        ),
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """Return a formula engine that outputs the minimum of `self` and `other`.

        Args:
            other: A formula receiver, a formula builder or a QuantityT instance
                corresponding to a sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method).min(other)

    def consumption(
        self: FormulaEngine3Phase[QuantityT],
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """
        Return a formula builder that applies the consumption operator on `self`.

        The consumption operator returns either the identity if the power value is
        positive or 0.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method).consumption()

    def production(
        self: FormulaEngine3Phase[QuantityT],
    ) -> HigherOrderFormulaBuilder3Phase[QuantityT]:
        """
        Return a formula builder that applies the production operator on `self`.

        The production operator returns either the absolute value if the power value is
        negative or 0.
        """
        return HigherOrderFormulaBuilder3Phase(self, self._create_method).production()

    async def _run(self) -> None:
        sender = self._channel.new_sender()
        phase_1_rx = self._streams[0].new_receiver()
        phase_2_rx = self._streams[1].new_receiver()
        phase_3_rx = self._streams[2].new_receiver()

        while True:
            try:
                phase_1 = await phase_1_rx.receive()
                phase_2 = await phase_2_rx.receive()
                phase_3 = await phase_3_rx.receive()
                msg = Sample3Phase(
                    phase_1.timestamp,
                    phase_1.value,
                    phase_2.value,
                    phase_3.value,
                )
            except asyncio.CancelledError:
                _logger.exception("FormulaEngine task cancelled: %s", self._name)
                break
            else:
                await sender.send(msg)

    def new_receiver(
        self, name: str | None = None, max_size: int = 50
    ) -> Receiver[Sample3Phase[QuantityT]]:
        """Create a new receiver that streams the output of the formula engine.

        Args:
            name: An optional name for the receiver.
            max_size: The size of the receiver's buffer.

        Returns:
            A receiver that streams output `Sample`s from the formula engine.
        """
        if self._task is None:
            self._task = asyncio.create_task(self._run())

        recv = self._channel.new_receiver(name=name, limit=max_size)

        # This is a hack to ensure that the lifetime of the engine is tied to the
        # lifetime of the receiver.  This is necessary because the engine is a task that
        # runs forever, and in cases where higher order built for example with the below
        # idiom, the user would hold no references to the engine and it could get
        # garbage collected before the receiver.  This behaviour is explained in the
        # `asyncio.create_task` docs here:
        # https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
        #
        #     formula = (grid_power_engine + bat_power_engine).build().new_receiver()
        recv._engine_reference = self  # type: ignore # pylint: disable=protected-access
        return recv


class FormulaBuilder(Generic[QuantityT]):
    """Builds a post-fix formula engine that operates on `Sample` receivers.

    Operators and metrics need to be pushed in in-fix order, and they get rearranged
    into post-fix order.  This is done using the [Shunting yard
    algorithm](https://en.wikipedia.org/wiki/Shunting_yard_algorithm).

    Example:
        To create an engine that adds the latest entries from two receivers, the
        following calls need to be made:

        ```python
        from frequenz.quantities import Power

        channel = Broadcast[Sample[Power]](name="channel")
        receiver_1 = channel.new_receiver(name="receiver_1")
        receiver_2 = channel.new_receiver(name="receiver_2")
        builder = FormulaBuilder("addition", Power)
        builder.push_metric("metric_1", receiver_1, nones_are_zeros=True)
        builder.push_oper("+")
        builder.push_metric("metric_2", receiver_2, nones_are_zeros=True)
        engine = builder.build()
        ```

        and then every call to `engine.apply()` would fetch a value from each receiver,
        add the values and return the result.
    """

    def __init__(self, name: str, create_method: Callable[[float], QuantityT]) -> None:
        """Create a `FormulaBuilder` instance.

        Args:
            name: A name for the formula being built.
            create_method: A method to generate the output `Sample` value with.  If the
                formula is for generating power values, this would be
                `Power.from_watts`, for example.
        """
        self._name = name
        self._create_method: Callable[[float], QuantityT] = create_method
        self._build_stack: list[FormulaStep] = []
        self._steps: list[FormulaStep] = []
        self._metric_fetchers: dict[str, MetricFetcher[QuantityT]] = {}

    def push_oper(self, oper: str) -> None:  # pylint: disable=too-many-branches
        """Push an operator into the engine.

        Args:
            oper: One of these strings - "+", "-", "*", "/", "(", ")"
        """
        if self._build_stack and oper != "(":
            op_prec = _operator_precedence[oper]
            while self._build_stack:
                prev_step = self._build_stack[-1]
                if op_prec < _operator_precedence[repr(prev_step)]:
                    break
                if oper == ")" and repr(prev_step) == "(":
                    self._build_stack.pop()
                    break
                if repr(prev_step) == "(":
                    break
                self._steps.append(prev_step)
                self._build_stack.pop()

        if oper == "+":
            self._build_stack.append(Adder())
        elif oper == "-":
            self._build_stack.append(Subtractor())
        elif oper == "*":
            self._build_stack.append(Multiplier())
        elif oper == "/":
            self._build_stack.append(Divider())
        elif oper == "(":
            self._build_stack.append(OpenParen())
        elif oper == "max":
            self._build_stack.append(Maximizer())
        elif oper == "min":
            self._build_stack.append(Minimizer())
        elif oper == "consumption":
            self._build_stack.append(Consumption())
        elif oper == "production":
            self._build_stack.append(Production())

    def push_metric(
        self,
        name: str,
        data_stream: Receiver[Sample[QuantityT]],
        *,
        nones_are_zeros: bool,
        fallback: FallbackMetricFetcher[QuantityT] | None = None,
    ) -> None:
        """Push a metric receiver into the engine.

        Args:
            name: A name for the metric.
            data_stream: A receiver to fetch this metric from.
            nones_are_zeros: Whether to treat None values from the stream as 0s.  If
                False, the returned value will be a None.
            fallback: Metric fetcher to use if primary one start sending
                invalid data (e.g. due to a component stop). If None, the data from
                primary metric fetcher will be used.
        """
        fetcher = self._metric_fetchers.setdefault(
            name,
            MetricFetcher(
                name,
                data_stream,
                nones_are_zeros=nones_are_zeros,
                fallback=fallback,
            ),
        )
        self._steps.append(fetcher)

    def push_constant(self, value: float) -> None:
        """Push a constant value into the engine.

        Args:
            value: The constant value to push.
        """
        self._steps.append(ConstantValue(value))

    def push_clipper(self, min_value: float | None, max_value: float | None) -> None:
        """Push a clipper step into the engine.

        The clip will be applied on the last value available on the evaluation stack,
        before the clip step is called.

        So if an entire expression needs to be clipped, the expression should be
        enclosed in parentheses, before the clip step is added.

        For example, this clips the output of the entire expression:

        ```python
        from frequenz.quantities import Power

        builder = FormulaBuilder("example", Power)
        channel = Broadcast[Sample[Power]](name="channel")
        receiver_1 = channel.new_receiver(name="receiver_1")
        receiver_2 = channel.new_receiver(name="receiver_2")

        builder.push_oper("(")
        builder.push_metric("metric_1", receiver_1, nones_are_zeros=True)
        builder.push_oper("+")
        builder.push_metric("metric_2", receiver_2, nones_are_zeros=True)
        builder.push_oper(")")
        builder.push_clipper(min_value=0.0, max_value=None)
        ```

        And this clips the output of metric_2 only, and not the final result:

        ```python
        from frequenz.quantities import Power

        builder = FormulaBuilder("example", Power)
        channel = Broadcast[Sample[Power]](name="channel")
        receiver_1 = channel.new_receiver(name="receiver_1")
        receiver_2 = channel.new_receiver(name="receiver_2")

        builder.push_metric("metric_1", receiver_1, nones_are_zeros=True)
        builder.push_oper("+")
        builder.push_metric("metric_2", receiver_2, nones_are_zeros=True)
        builder.push_clipper(min_value=0.0, max_value=None)
        ```

        Args:
            min_value: The minimum value to clip to.
            max_value: The maximum value to clip to.
        """
        self._steps.append(Clipper(min_value, max_value))

    @property
    def name(self) -> str:
        """Return the name of the formula being built.

        Returns:
            The name of the formula being built.
        """
        return self._name

    async def subscribe(self) -> None:
        """Subscribe to metrics if needed.

        This is a no-op for the `FormulaBuilder` class, but is used by the
        `ResampledFormulaBuilder` class.
        """

    def finalize(
        self,
    ) -> tuple[list[FormulaStep], dict[str, MetricFetcher[QuantityT]]]:
        """Finalize and return the steps and fetchers for the formula.

        Returns:
            A tuple of the steps and fetchers for the formula.
        """
        while self._build_stack:
            self._steps.append(self._build_stack.pop())

        return self._steps, self._metric_fetchers

    def __str__(self) -> str:
        """Return a string representation of the formula.

        Returns:
            A string representation of the formula.
        """
        steps = self._steps if len(self._steps) > 0 else self._build_stack
        return format_formula(steps)

    def build(self) -> FormulaEngine[QuantityT]:
        """Create a formula engine with the steps and fetchers that have been pushed.

        Returns:
            A `FormulaEngine` instance.
        """
        self.finalize()
        return FormulaEngine(self, create_method=self._create_method)


FormulaEngineT = TypeVar(
    "FormulaEngineT", bound=(FormulaEngine[Any] | FormulaEngine3Phase[Any])
)


class _BaseHOFormulaBuilder(ABC, Generic[FormulaEngineT, QuantityT]):
    """Provides a way to build formulas from the outputs of other formulas."""

    def __init__(
        self,
        engine: FormulaEngineT,
        create_method: Callable[[float], QuantityT],
    ) -> None:
        """Create a `GenericHigherOrderFormulaBuilder` instance.

        Args:
            engine: A first input stream to create a builder with, so that python
                operators `+, -, *, /` can be used directly on newly created instances.
            create_method: A method to generate the output `Sample` value with.  If the
                formula is for generating power values, this would be
                `Power.from_watts`, for example.
        """
        self._steps: deque[
            tuple[
                TokenType,
                FormulaEngine[QuantityT]
                | FormulaEngine3Phase[QuantityT]
                | Quantity
                | float
                | str,
            ]
        ] = deque()
        self._steps.append((TokenType.COMPONENT_METRIC, engine))
        self._create_method: Callable[[float], QuantityT] = create_method

    def _push(
        self,
        oper: str,
        other: Self | FormulaEngineT | QuantityT | float,
    ) -> Self:
        self._steps.appendleft((TokenType.OPER, "("))
        self._steps.append((TokenType.OPER, ")"))
        self._steps.append((TokenType.OPER, oper))

        if isinstance(other, (FormulaEngine, FormulaEngine3Phase)):
            self._steps.append((TokenType.COMPONENT_METRIC, other))
        elif isinstance(other, (Quantity, float, int)):
            match oper:
                case "+" | "-" | "max" | "min":
                    if not isinstance(other, Quantity):
                        raise RuntimeError(
                            "A Quantity must be provided for addition,"
                            f" subtraction, min or max to {other}"
                        )
                case "*" | "/":
                    if not isinstance(other, (float, int)):
                        raise RuntimeError(
                            f"A float must be provided for scalar multiplication to {other}"
                        )
            self._steps.append((TokenType.CONSTANT, other))
        elif isinstance(other, _BaseHOFormulaBuilder):
            self._steps.append((TokenType.OPER, "("))
            self._steps.extend(other._steps)  # pylint: disable=protected-access
            self._steps.append((TokenType.OPER, ")"))
        else:
            raise RuntimeError(f"Can't build a formula from: {other}")
        return self

    def __add__(
        self,
        other: Self | FormulaEngineT | QuantityT,
    ) -> Self:
        """Return a formula builder that adds (data in) `other` to `self`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return self._push("+", other)

    def __sub__(
        self,
        other: Self | FormulaEngineT | QuantityT,
    ) -> Self:
        """Return a formula builder that subtracts (data in) `other` from `self`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return self._push("-", other)

    def __mul__(
        self,
        other: Self | FormulaEngineT | float,
    ) -> Self:
        """Return a formula builder that multiplies (data in) `self` with `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return self._push("*", other)

    def __truediv__(
        self,
        other: Self | FormulaEngineT | float,
    ) -> Self:
        """Return a formula builder that divides (data in) `self` by `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return self._push("/", other)

    def max(
        self,
        other: Self | FormulaEngineT | QuantityT,
    ) -> Self:
        """Return a formula builder that calculates the maximum of `self` and `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return self._push("max", other)

    def min(
        self,
        other: Self | FormulaEngineT | QuantityT,
    ) -> Self:
        """Return a formula builder that calculates the minimum of `self` and `other`.

        Args:
            other: A formula receiver, or a formula builder instance corresponding to a
                sub-expression.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        return self._push("min", other)

    def consumption(
        self,
    ) -> Self:
        """Apply the Consumption Operator.

        The consumption operator returns either the identity if the power value is
        positive or 0.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        self._steps.appendleft((TokenType.OPER, "("))
        self._steps.append((TokenType.OPER, ")"))
        self._steps.append((TokenType.OPER, "consumption"))
        return self

    def production(
        self,
    ) -> Self:
        """Apply the Production Operator.

        The production operator returns either the absolute value if the power value is
        negative or 0.

        Returns:
            A formula builder that can take further expressions, or can be built
                into a formula engine.
        """
        self._steps.appendleft((TokenType.OPER, "("))
        self._steps.append((TokenType.OPER, ")"))
        self._steps.append((TokenType.OPER, "production"))
        return self


class HigherOrderFormulaBuilder(
    Generic[QuantityT], _BaseHOFormulaBuilder[FormulaEngine[QuantityT], QuantityT]
):
    """A specialization of the _BaseHOFormulaBuilder for `FormulaReceiver`."""

    def build(
        self, name: str, *, nones_are_zeros: bool = False
    ) -> FormulaEngine[QuantityT]:
        """Build a `FormulaEngine` instance from the builder.

        Args:
            name: A name for the newly generated formula.
            nones_are_zeros: whether `None` values in any of the input streams should be
                treated as zeros.

        Returns:
            A `FormulaEngine` instance.
        """
        builder = FormulaBuilder(name, self._create_method)
        for typ, value in self._steps:
            if typ == TokenType.COMPONENT_METRIC:
                assert isinstance(value, FormulaEngine)
                builder.push_metric(
                    value._name,  # pylint: disable=protected-access
                    value.new_receiver(),
                    nones_are_zeros=nones_are_zeros,
                )
            elif typ == TokenType.OPER:
                assert isinstance(value, str)
                builder.push_oper(value)
            elif typ == TokenType.CONSTANT:
                assert isinstance(value, (Quantity, float))
                builder.push_constant(
                    value.base_value if isinstance(value, Quantity) else value
                )
        return builder.build()


class HigherOrderFormulaBuilder3Phase(
    Generic[QuantityT], _BaseHOFormulaBuilder[FormulaEngine3Phase[QuantityT], QuantityT]
):
    """A specialization of the _BaseHOFormulaBuilder for `FormulaReceiver3Phase`."""

    def build(
        self, name: str, *, nones_are_zeros: bool = False
    ) -> FormulaEngine3Phase[QuantityT]:
        """Build a `FormulaEngine3Phase` instance from the builder.

        Args:
            name: A name for the newly generated formula.
            nones_are_zeros: whether `None` values in any of the input streams should be
                treated as zeros.

        Returns:
            A `FormulaEngine3Phase` instance.
        """
        builders = [
            FormulaBuilder(name, self._create_method),
            FormulaBuilder(name, self._create_method),
            FormulaBuilder(name, self._create_method),
        ]
        for typ, value in self._steps:
            if typ == TokenType.COMPONENT_METRIC:
                assert isinstance(value, FormulaEngine3Phase)
                for phase in range(3):
                    builders[phase].push_metric(
                        f"{value._name}-{phase+1}",  # pylint: disable=protected-access
                        value._streams[  # pylint: disable=protected-access
                            phase
                        ].new_receiver(),
                        nones_are_zeros=nones_are_zeros,
                    )
            elif typ == TokenType.OPER:
                assert isinstance(value, str)
                for phase in range(3):
                    builders[phase].push_oper(value)
        return FormulaEngine3Phase(
            name,
            self._create_method,
            (
                builders[0].build(),
                builders[1].build(),
                builders[2].build(),
            ),
        )



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Wrapper around the power managing and power distributing actors."""

from __future__ import annotations

import logging
from datetime import timedelta

from frequenz.channels import Broadcast

# pylint seems to think this is a cyclic import, but it is not.
#
# pylint: disable=cyclic-import
from frequenz.client.microgrid import ComponentCategory, ComponentType

from .._internal._channels import ChannelRegistry, ReceiverFetcher
from . import _power_managing, connection_manager
from ._power_distributing import (
    ComponentPoolStatus,
    PowerDistributingActor,
    Request,
    Result,
)

_logger = logging.getLogger(__name__)


class PowerWrapper:
    """Wrapper around the power managing and power distributing actors."""

    def __init__(
        self,
        channel_registry: ChannelRegistry,
        *,
        api_power_request_timeout: timedelta,
        component_category: ComponentCategory,
        component_type: ComponentType | None = None,
    ):
        """Initialize the power control.

        Args:
            channel_registry: A channel registry for use in the actors.
            api_power_request_timeout: Timeout to use when making power requests to
                the microgrid API.
            component_category: The category of the components that actors started by
                this instance of the PowerWrapper will be responsible for.
            component_type: The type of the component of the given category that this
                actor is responsible for.  This is used only when the component category
                is not enough to uniquely identify the component.  For example, when the
                category is `ComponentCategory.INVERTER`, the type is needed to identify
                the inverter as a solar inverter or a battery inverter.  This can be
                `None` when the component category is enough to uniquely identify the
                component.
        """
        self._component_category = component_category
        self._component_type = component_type
        self._channel_registry = channel_registry
        self._api_power_request_timeout = api_power_request_timeout

        self.status_channel: Broadcast[ComponentPoolStatus] = Broadcast(
            name="Component Status Channel", resend_latest=True
        )
        self._power_distribution_requests_channel: Broadcast[Request] = Broadcast(
            name="Power Distributing Actor, Requests Broadcast Channel"
        )
        self._power_distribution_results_channel: Broadcast[Result] = Broadcast(
            name="Power Distributing Actor, Results Broadcast Channel"
        )

        self.proposal_channel: Broadcast[_power_managing.Proposal] = Broadcast(
            name="Power Managing Actor, Requests Broadcast Channel"
        )
        self.bounds_subscription_channel: Broadcast[_power_managing.ReportRequest] = (
            Broadcast(name="Power Managing Actor, Bounds Subscription Channel")
        )

        self._power_distributing_actor: PowerDistributingActor | None = None
        self._power_managing_actor: _power_managing.PowerManagingActor | None = None

    def _start_power_managing_actor(self) -> None:
        """Start the power managing actor if it is not already running."""
        if self._power_managing_actor:
            return

        component_graph = connection_manager.get().component_graph
        # Currently the power managing actor only supports batteries.  The below
        # constraint needs to be relaxed if the actor is extended to support other
        # components.
        if not component_graph.components(
            component_categories={self._component_category}
        ):
            _logger.warning(
                "No %s found in the component graph. "
                "The power managing actor will not be started.",
                self._component_category,
            )
            return

        self._power_managing_actor = _power_managing.PowerManagingActor(
            component_category=self._component_category,
            component_type=self._component_type,
            proposals_receiver=self.proposal_channel.new_receiver(),
            bounds_subscription_receiver=(
                self.bounds_subscription_channel.new_receiver()
            ),
            power_distributing_requests_sender=(
                self._power_distribution_requests_channel.new_sender()
            ),
            power_distributing_results_receiver=(
                self._power_distribution_results_channel.new_receiver()
            ),
            channel_registry=self._channel_registry,
        )
        self._power_managing_actor.start()

    def _start_power_distributing_actor(self) -> None:
        """Start the power distributing actor if it is not already running."""
        if self._power_distributing_actor:
            return

        component_graph = connection_manager.get().component_graph
        if not component_graph.components(
            component_categories={self._component_category}
        ):
            _logger.warning(
                "No %s found in the component graph. "
                "The power distributing actor will not be started.",
                self._component_category,
            )
            return

        # The PowerDistributingActor is started with only a single default user channel.
        # Until the PowerManager is implemented, support for multiple use-case actors
        # will not be available in the high level interface.
        self._power_distributing_actor = PowerDistributingActor(
            component_category=self._component_category,
            component_type=self._component_type,
            api_power_request_timeout=self._api_power_request_timeout,
            requests_receiver=self._power_distribution_requests_channel.new_receiver(),
            results_sender=self._power_distribution_results_channel.new_sender(),
            component_pool_status_sender=self.status_channel.new_sender(),
        )
        self._power_distributing_actor.start()

    @property
    def started(self) -> bool:
        """Return True if power managing and power distributing actors are started."""
        return (
            self._power_managing_actor is not None
            and self._power_distributing_actor is not None
        )

    def start(self) -> None:
        """Start the power managing and power distributing actors."""
        if self.started:
            return
        self._start_power_distributing_actor()
        self._start_power_managing_actor()

    async def stop(self) -> None:
        """Stop the power managing and power distributing actors."""
        if self._power_distributing_actor:
            await self._power_distributing_actor.stop()
        if self._power_managing_actor:
            await self._power_managing_actor.stop()

    def distribution_results_fetcher(self) -> ReceiverFetcher[Result]:
        """Return a fetcher for the power distribution results."""
        return self._power_distribution_results_channel



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Interactions with PV inverters."""

from ._pv_pool import PVPool, PVPoolError
from ._result_types import PVPoolReport

__all__ = [
    "PVPool",
    "PVPoolError",
    "PVPoolReport",
]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""System bounds tracker for PV inverters."""

import asyncio
from collections import abc

from frequenz.channels import Receiver, Sender, merge, select, selected_from
from frequenz.client.microgrid import InverterData
from frequenz.quantities import Power

from ..._internal._asyncio import run_forever
from ...actor import BackgroundService
from ...microgrid import connection_manager
from ...microgrid._power_distributing._component_status import ComponentPoolStatus
from .._base_types import Bounds, SystemBounds


class PVSystemBoundsTracker(BackgroundService):
    """Track the system bounds for PV inverters.

    System bounds are the aggregate bounds for the PV inverters in the pool that are in
    a working state.  They are calculated from the individual bounds received from the
    microgrid API.

    The system bounds are sent to the `bounds_sender` whenever they change.
    """

    def __init__(
        self,
        component_ids: abc.Set[int],
        status_receiver: Receiver[ComponentPoolStatus],
        bounds_sender: Sender[SystemBounds],
    ):
        """Initialize the system bounds tracker.

        Args:
            component_ids: The ids of the components to track.
            status_receiver: A receiver that streams the status of the PV inverters in
                the pool.
            bounds_sender: A sender to send the system bounds to.
        """
        super().__init__()

        self._component_ids = component_ids
        self._status_receiver = status_receiver
        self._bounds_sender = bounds_sender
        self._latest_component_data: dict[int, InverterData] = {}
        self._last_sent_bounds: SystemBounds | None = None
        self._component_pool_status = ComponentPoolStatus(set(), set())

    def start(self) -> None:
        """Start the PV inverter system bounds tracker."""
        self._tasks.add(asyncio.create_task(run_forever(self._run)))

    async def _send_bounds(self) -> None:
        """Calculate and send the aggregate system bounds if they have changed."""
        if not self._latest_component_data:
            return
        inclusion_bounds = Bounds(
            lower=Power.from_watts(
                sum(
                    data.active_power_inclusion_lower_bound
                    for data in self._latest_component_data.values()
                )
            ),
            upper=Power.from_watts(
                sum(
                    data.active_power_inclusion_upper_bound
                    for data in self._latest_component_data.values()
                )
            ),
        )
        exclusion_bounds = Bounds(
            lower=Power.from_watts(
                sum(
                    data.active_power_exclusion_lower_bound
                    for data in self._latest_component_data.values()
                )
            ),
            upper=Power.from_watts(
                sum(
                    data.active_power_exclusion_upper_bound
                    for data in self._latest_component_data.values()
                )
            ),
        )

        if (
            self._last_sent_bounds is None
            or self._last_sent_bounds.inclusion_bounds != inclusion_bounds
            or self._last_sent_bounds.exclusion_bounds != exclusion_bounds
        ):
            self._last_sent_bounds = SystemBounds(
                timestamp=max(
                    data.timestamp for data in self._latest_component_data.values()
                ),
                inclusion_bounds=inclusion_bounds,
                exclusion_bounds=exclusion_bounds,
            )
            await self._bounds_sender.send(self._last_sent_bounds)

    async def _run(self) -> None:
        """Run the system bounds tracker."""
        api_client = connection_manager.get().api_client
        status_rx = self._status_receiver
        pv_data_rx = merge(
            *(
                await asyncio.gather(
                    *(
                        api_client.inverter_data(component_id)
                        for component_id in self._component_ids
                    )
                )
            )
        )

        async for selected in select(status_rx, pv_data_rx):
            if selected_from(selected, status_rx):
                self._component_pool_status = selected.message
                to_remove = []
                for comp_id in self._latest_component_data:
                    if (
                        comp_id not in self._component_pool_status.working
                        and comp_id not in self._component_pool_status.uncertain
                    ):
                        to_remove.append(comp_id)
                for comp_id in to_remove:
                    del self._latest_component_data[comp_id]
            elif selected_from(selected, pv_data_rx):
                data = selected.message
                comp_id = data.component_id
                if (
                    comp_id not in self._component_pool_status.working
                    and comp_id not in self._component_pool_status.uncertain
                ):
                    continue
                self._latest_component_data[data.component_id] = data

            await self._send_bounds()



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""
Handling of timeseries streams.

A timeseries is a stream (normally an async iterator) of
[`Sample`][frequenz.sdk.timeseries.Sample]s.

# Periodicity and alignment

All the data produced by this package is always periodic and aligned to the
`UNIX_EPOCH` (by default).

Classes normally take a (re)sampling period as and argument and, optionally, an
`align_to` argument.

This means timestamps are always separated exactly by a period, and that this
timestamp falls always at multiples of the period, starting at the `align_to`.

This ensures that the data is predictable and consistent among restarts.

Example:
    If we have a period of 10 seconds, and are aligning to the UNIX
    epoch. Assuming the following timeline starts in 1970-01-01 00:00:00
    UTC and our current `now` is 1970-01-01 00:00:32 UTC, then the next
    timestamp will be at 1970-01-01 00:00:40 UTC:

    ```
    align_to = 1970-01-01 00:00:00         next event = 1970-01-01 00:00:40
    |                                       |
    |---------|---------|---------|-|-------|---------|---------|---------|
    0        10        20        30 |      40        50        60        70
                                   now = 1970-01-01 00:00:32
    ```
"""

from .._internal._channels import ReceiverFetcher
from ._base_types import UNIX_EPOCH, Bounds, Sample, Sample3Phase
from ._fuse import Fuse
from ._moving_window import MovingWindow
from ._periodic_feature_extractor import PeriodicFeatureExtractor
from ._resampling import ResamplerConfig

__all__ = [
    "Bounds",
    "Fuse",
    "MovingWindow",
    "PeriodicFeatureExtractor",
    "ResamplerConfig",
    "ReceiverFetcher",
    "Sample",
    "Sample3Phase",
    "UNIX_EPOCH",
]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Module to define how to subscribe and fetch component data."""

from __future__ import annotations

import asyncio
import logging
import math
from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import datetime, timezone
from typing import Any, Generic, Self, TypeVar

from frequenz.channels import ChannelClosedError, Receiver
from frequenz.client.microgrid import (
    BatteryData,
    ComponentCategory,
    ComponentData,
    ComponentMetricId,
    InverterData,
)

from ..._internal._asyncio import AsyncConstructible
from ..._internal._constants import MAX_BATTERY_DATA_AGE_SEC
from ...microgrid import connection_manager
from ...microgrid._data_sourcing.microgrid_api_source import (
    _BatteryDataMethods,
    _InverterDataMethods,
)
from ._component_metrics import ComponentMetricsData

_logger = logging.getLogger(__name__)

T = TypeVar("T", bound=ComponentData)
"""Type variable for component data."""


class ComponentMetricFetcher(AsyncConstructible, ABC):
    """Define how to subscribe for and fetch the component metrics data."""

    _component_id: int
    _metrics: Iterable[ComponentMetricId]

    @classmethod
    async def async_new(
        cls, component_id: int, metrics: Iterable[ComponentMetricId]
    ) -> Self:
        """Create an instance of this class.

        Subscribe for the given component metrics and return them if method
        `fetch_next` is called.

        Args:
            component_id: component id
            metrics: metrics that should be fetched from this component.

        Returns:
            This class instance.
        """
        self: Self = cls.__new__(cls)
        self._component_id = component_id
        self._metrics = metrics
        return self

    @abstractmethod
    async def fetch_next(self) -> ComponentMetricsData | None:
        """Fetch metrics for this component."""


class LatestMetricsFetcher(ComponentMetricFetcher, Generic[T], ABC):
    """Subscribe for the latest component data and extract the needed metrics."""

    _receiver: Receiver[T]
    _max_waiting_time: float

    @classmethod
    async def async_new(
        cls,
        component_id: int,
        metrics: Iterable[ComponentMetricId],
    ) -> Self:
        """Create instance of this class.

        Subscribe for the requested component data and fetch only the latest component
        metrics.

        Args:
            component_id: component id
            metrics: metrics

        Raises:
            ValueError: If any requested metric id is not supported.

        Returns:
            This class instance
        """
        self: Self = await super().async_new(component_id, metrics)

        for metric in metrics:
            # pylint: disable=protected-access
            if metric not in self._supported_metrics():
                category = self._component_category()
                raise ValueError(f"Metric {metric} not supported for {category}")

        # pylint: disable=protected-access
        self._receiver = await self._subscribe()
        self._max_waiting_time = MAX_BATTERY_DATA_AGE_SEC
        return self

    async def fetch_next(self) -> ComponentMetricsData | None:
        """Fetch the latest component metrics.

        Returns:
            Component metrics data.
            None if the channel was closed and fetching next element is impossible.
        """
        try:
            data = await asyncio.wait_for(
                self._receiver.receive(), self._max_waiting_time
            )

        except ChannelClosedError:
            _logger.exception(
                "Channel for component %d was closed.", self._component_id
            )
            return None
        except asyncio.TimeoutError:
            # Next time wait infinitely until we receive any message.
            _logger.debug("Component %d stopped sending data.", self._component_id)
            return ComponentMetricsData(
                self._component_id, datetime.now(tz=timezone.utc), {}
            )

        self._max_waiting_time = MAX_BATTERY_DATA_AGE_SEC
        metrics = {}
        for mid in self._metrics:
            value = self._extract_metric(data, mid)
            # There is no guarantee that all fields in component message are populated
            if not math.isnan(value):
                metrics[mid] = value

        return ComponentMetricsData(self._component_id, data.timestamp, metrics)

    @abstractmethod
    def _extract_metric(self, data: T, mid: ComponentMetricId) -> float: ...

    @abstractmethod
    def _supported_metrics(self) -> set[ComponentMetricId]: ...

    @abstractmethod
    def _component_category(self) -> ComponentCategory: ...

    @abstractmethod
    async def _subscribe(self) -> Receiver[Any]:
        """Subscribe for this component data.

        Size of the receiver buffer should should be 1 to make sure we receive only
        the latest component data.

        Returns:
            Receiver for this component metrics.
        """


class LatestBatteryMetricsFetcher(LatestMetricsFetcher[BatteryData]):
    """Subscribe for the latest battery data using MicrogridApiClient."""

    @classmethod
    async def async_new(  # noqa: DOC502 (ValueError is raised indirectly super.async_new)
        cls,
        component_id: int,
        metrics: Iterable[ComponentMetricId],
    ) -> LatestBatteryMetricsFetcher:
        """Create instance of this class.

        Subscribe for the requested component data and fetch only the latest component
        metrics.

        Args:
            component_id: component id
            metrics: metrics

        Raises:
            ValueError: If any requested metric id is not supported.

        Returns:
            This class instance
        """
        self: LatestBatteryMetricsFetcher = await super().async_new(
            component_id, metrics
        )
        return self

    def _supported_metrics(self) -> set[ComponentMetricId]:
        return set(_BatteryDataMethods.keys())

    def _extract_metric(self, data: BatteryData, mid: ComponentMetricId) -> float:
        return _BatteryDataMethods[mid](data)

    async def _subscribe(self) -> Receiver[BatteryData]:
        """Subscribe for this component data.

        Size of the receiver buffer should should be 1 to make sure we receive only
        the latest component data.

        Returns:
            Receiver for this component metrics.
        """
        api = connection_manager.get().api_client
        return await api.battery_data(self._component_id, maxsize=1)

    def _component_category(self) -> ComponentCategory:
        return ComponentCategory.BATTERY


class LatestInverterMetricsFetcher(LatestMetricsFetcher[InverterData]):
    """Subscribe for the latest inverter data using MicrogridApiClient."""

    @classmethod
    async def async_new(  # noqa: DOC502 (ValueError is raised indirectly by super.async_new)
        cls,
        component_id: int,
        metrics: Iterable[ComponentMetricId],
    ) -> LatestInverterMetricsFetcher:
        """Create instance of this class.

        Subscribe for the requested component data and fetch only the latest component
        metrics.

        Args:
            component_id: component id
            metrics: metrics

        Raises:
            ValueError: If any requested metric id is not supported.

        Returns:
            This class instance
        """
        self: LatestInverterMetricsFetcher = await super().async_new(
            component_id, metrics
        )
        return self

    def _supported_metrics(self) -> set[ComponentMetricId]:
        return set(_InverterDataMethods.keys())

    def _extract_metric(self, data: InverterData, mid: ComponentMetricId) -> float:
        return _InverterDataMethods[mid](data)

    async def _subscribe(self) -> Receiver[InverterData]:
        """Subscribe for this component data.

        Size of the receiver buffer should should be 1 to make sure we receive only
        the latest component data.

        Returns:
            Receiver for this component metrics.
        """
        api = connection_manager.get().api_client
        return await api.inverter_data(self._component_id, maxsize=1)

    def _component_category(self) -> ComponentCategory:
        return ComponentCategory.INVERTER



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Definition of Singleton metaclass."""

from threading import Lock
from typing import Any


class SingletonMeta(type):
    """This is a thread-safe implementation of Singleton."""

    _instances: dict[Any, type] = {}
    """The dictionary of instances of the singleton classes."""

    _lock: Lock = Lock()
    """The lock to ensure thread safety.

    The lock to acquire when creating a singleton instance, preventing
    multiple threads from creating instances simultaneously.
    """

    def __call__(cls, *args: Any, **kwargs: Any) -> type:
        """Overload function call operator to return the singleton instance.

        Args:
            *args: positional args
            **kwargs: and keyword args for the super class.

        Returns:
            The singleton instance.
        """
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Classes to track the status of components in the microgrid."""


import enum
from abc import ABC, abstractmethod
from collections import abc
from dataclasses import dataclass
from datetime import timedelta

from frequenz.channels import Receiver, Sender

from ....actor._background_service import BackgroundService


@dataclass
class ComponentPoolStatus:
    """Status of all components of a certain category in the microgrid."""

    working: set[int]
    """Set of working component ids."""

    uncertain: set[int]
    """Set of components to be used only when there are none known to be working."""

    def get_working_components(self, components: abc.Set[int]) -> set[int]:
        """From the given set of components return the working ones.

        Args:
            components: Set of components.

        Returns:
            Subset with working components.
        """
        working = self.working.intersection(components)
        if len(working) > 0:
            return working
        return self.uncertain.intersection(components)


class ComponentStatusEnum(enum.Enum):
    """Enum for component status."""

    NOT_WORKING = 0
    """Component is not working and should not be used."""

    UNCERTAIN = 1
    """Component should work, although the last request to it failed.

    It is blocked for few seconds and it is not recommended to use it unless it is
    necessary.
    """

    WORKING = 2
    """Component is working"""


@dataclass(frozen=True)
class ComponentStatus:
    """Status of a single component."""

    component_id: int
    """Component ID."""

    value: ComponentStatusEnum
    """Component status."""


@dataclass(frozen=True, kw_only=True)
class SetPowerResult:
    """Lists of components for which the last set power command succeeded or failed."""

    succeeded: abc.Set[int]
    """Component IDs for which the last set power command succeeded."""

    failed: abc.Set[int]
    """Component IDs for which the last set power command failed."""


class ComponentStatusTracker(BackgroundService, ABC):
    """Interface for specialized component status trackers to implement."""

    @abstractmethod
    def __init__(  # pylint: disable=too-many-arguments,super-init-not-called
        self,
        *,
        component_id: int,
        max_data_age: timedelta,
        max_blocking_duration: timedelta,
        status_sender: Sender[ComponentStatus],
        set_power_result_receiver: Receiver[SetPowerResult],
    ) -> None:
        """Create class instance.

        Args:
            component_id: Id of this component
            max_data_age: If component stopped sending data, then this is the maximum
                time when its last message should be considered as valid. After that
                time, component won't be used until it starts sending data.
            max_blocking_duration: This value tell what should be the maximum
                timeout used for blocking failing component.
            status_sender: Channel to send status updates.
            set_power_result_receiver: Channel to receive results of the requests to the
                components.
        """



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for 3-phase Grid Current."""

from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Current

from .._formula_engine import FormulaEngine, FormulaEngine3Phase
from ._formula_generator import FormulaGenerator


class GridCurrentFormula(FormulaGenerator[Current]):
    """Create a formula engine from the component graph for calculating grid current."""

    def generate(  # noqa: DOC502
        # ComponentNotFound is raised indirectly by _get_grid_component_successors
        self,
    ) -> FormulaEngine3Phase[Current]:
        """Generate a formula for calculating grid current from the component graph.

        Returns:
            A formula engine that will calculate 3-phase grid current values.

        Raises:
            ComponentNotFound: when the component graph doesn't have a `GRID` component.
        """
        grid_successors = self._get_grid_component_successors()

        return FormulaEngine3Phase(
            "grid-current",
            Current.from_amperes,
            (
                self._gen_phase_formula(
                    grid_successors, ComponentMetricId.CURRENT_PHASE_1
                ),
                self._gen_phase_formula(
                    grid_successors, ComponentMetricId.CURRENT_PHASE_2
                ),
                self._gen_phase_formula(
                    grid_successors, ComponentMetricId.CURRENT_PHASE_3
                ),
            ),
        )

    def _gen_phase_formula(
        self,
        grid_successors: set[Component],
        metric_id: ComponentMetricId,
    ) -> FormulaEngine[Current]:
        builder = self._get_builder("grid-current", metric_id, Current.from_amperes)

        # generate a formula that just adds values from all components that are
        # directly connected to the grid.
        for idx, comp in enumerate(grid_successors):
            # When inverters or ev chargers produce `None` samples, those
            # inverters are excluded from the calculation by treating their
            # `None` values as `0`s.
            #
            # This is not possible for Meters, so when they produce `None`
            # values, those values get propagated as the output.
            if comp.category in (
                ComponentCategory.INVERTER,
                ComponentCategory.EV_CHARGER,
            ):
                nones_are_zeros = True
            elif comp.category == ComponentCategory.METER:
                nones_are_zeros = False
            else:
                continue

            if idx > 0:
                builder.push_oper("+")

            builder.push_component_metric(
                comp.component_id, nones_are_zeros=nones_are_zeros
            )

        return builder.build()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for Grid Power."""

import logging

from frequenz.client.microgrid import ComponentMetricId
from frequenz.quantities import Power

from .._formula_engine import FormulaEngine
from ._formula_generator import NON_EXISTING_COMPONENT_ID, FormulaGenerator

_logger = logging.getLogger(__name__)


class EVChargerPowerFormula(FormulaGenerator[Power]):
    """Create a formula engine from the component graph for calculating grid power."""

    def generate(self) -> FormulaEngine[Power]:
        """Generate a formula for calculating total EV power for given component ids.

        Returns:
            A formula engine that calculates total EV Charger power values.
        """
        builder = self._get_builder(
            "ev-power", ComponentMetricId.ACTIVE_POWER, Power.from_watts
        )

        component_ids = self._config.component_ids
        if not component_ids:
            _logger.warning(
                "No EV Charger component IDs specified. "
                "Subscribing to the resampling actor with a non-existing "
                "component id, so that `0` values are sent from the formula."
            )
            # If there are no EV Chargers, we have to send 0 values as the same
            # frequency as the other streams. So we subscribe with a non-existing
            # component id, just to get a `None` message at the resampling interval.
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID, nones_are_zeros=True
            )
            return builder.build()

        for idx, component_id in enumerate(component_ids):
            if idx > 0:
                builder.push_oper("+")
            builder.push_component_metric(component_id, nones_are_zeros=True)

        return builder.build()



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""FallbackMetricFetcher implementation that uses formula generator."""

from frequenz.channels import Receiver

from ..._base_types import QuantityT, Sample
from .. import FormulaEngine
from .._formula_steps import FallbackMetricFetcher
from ._formula_generator import FormulaGenerator


# This is done as a separate module to avoid circular imports.
class FallbackFormulaMetricFetcher(FallbackMetricFetcher[QuantityT]):
    """A metric fetcher that uses a formula generator.

    The formula engine is generated lazily, meaning it is created only when
    the `start` or `fetch_next` method is called for the first time.
    Once the formula engine is initialized, it subscribes to its components
    and begins calculating and sending the formula results.
    """

    def __init__(self, formula_generator: FormulaGenerator[QuantityT]):
        """Create a `FallbackFormulaMetricFetcher` instance.

        Args:
            formula_generator: A formula generator that generates
                a formula engine with fallback components.
        """
        super().__init__()
        self._name = formula_generator.namespace
        self._formula_generator: FormulaGenerator[QuantityT] = formula_generator
        self._formula_engine: FormulaEngine[QuantityT] | None = None
        self._receiver: Receiver[Sample[QuantityT]] | None = None

    @property
    def name(self) -> str:
        """Get the name of the fetcher."""
        return self._name

    @property
    def is_running(self) -> bool:
        """Check whether the formula engine is running."""
        return self._receiver is not None

    def start(self) -> None:
        """Initialize the formula engine and start fetching samples."""
        engine = self._formula_generator.generate()
        # We need this assert because generate() can return a FormulaEngine
        # or FormulaEngine3Phase, but in this case we know it will return a
        # FormulaEngine. This helps to silence `mypy` and also to verify our
        # assumptions are still true at runtime
        assert isinstance(engine, FormulaEngine)
        self._formula_engine = engine
        self._receiver = self._formula_engine.new_receiver()

    async def ready(self) -> bool:
        """Wait until the receiver is ready with a message or an error.

        Once a call to `ready()` has finished, the message should be read with
        a call to `consume()` (`receive()` or iterated over).

        Returns:
            Whether the receiver is still active.
        """
        if self._receiver is None:
            self.start()

        assert self._receiver is not None
        return await self._receiver.ready()

    def consume(self) -> Sample[QuantityT]:
        """Return the latest message once `ready()` is complete."""
        assert (
            self._receiver is not None
        ), f"Fallback metric fetcher: {self.name} was not started"

        return self._receiver.consume()
